from web_ui import app
from flask import session
from datetime import datetime


# For calculating scores
epoch = datetime.utcfromtimestamp(0)
epoch_seconds = lambda dt: (dt - epoch).total_seconds() - 1356048000


def score(star_object):
    import random
    return random.random() * 100 - random.random() * 10


def get_active_persona():
    from nucleus.models import Persona
    """ Return the currently active persona or 0 if there is no controlled persona. """

    if 'active_persona' not in session or session['active_persona'] is None:
        controlled_personas = Persona.query.filter('sign_private != ""')

        if controlled_personas.first() is None:
            return ""
        else:
            session['active_persona'] = controlled_personas.first().id

    return session['active_persona']


def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1] in app.config['ALLOWED_EXTENSIONS']



from topicmodel import TopicModel
from web_ui import db, app
from nucleus.models import Persona, Oneup, Star, LinkPlanet
from sklearn.naive_bayes import GaussianNB
from helpers import get_site_content
from datetime import datetime
import pdb


class InterestModel(db.Model):

    __tablename__ = "interestmodel"
    id = db.Column(db.String(32), primary_key=True)
    persona_id = db.Column(db.String(32), db.ForeignKey('persona.id'))
    persona = db.relationship("Persona",
                              backref=db.backref('interestmodel'),
                              primaryjoin="Persona.id==InterestModel.persona_id")
    classifier = db.Column(db.PickleType())
    last_fit = db.Column(db.DateTime)
    last_prediction = db.Column(db.DateTime)

    def __init__(self, persona_id):
        self.persona_id = persona_id
        self.classifier = GaussianNB()

        self.last_fit = datetime.now()
        self.last_prediction = 0

    def is_interesting(self, text):
        topics = topic_model.get_topics_text(text)
        interesting = self.classifier.predict(topics)
        return interesting


def update():
    app.logger.info("Updating interest model")

    topic_model = TopicModel(
        app.config["TOPIC_MODEL"], app.config["TOPIC_MODEL_IDS"])

    for persona in Persona.query.all():
        interestmodel = InterestModel.query.filter_by(persona_id=persona.id).first()

        if interestmodel is None:
            interestmodel = InterestModel(persona.id)

        fit(interestmodel, topic_model)

    del topic_model
    app.logger.info("Update finished")


def fit(interestmodel, topic_model):
    train_set_pos = []
    train_set_neg = []
    for star in Star.query.filter_by(state=0):
        like = star.author_id == interestmodel.persona_id
        if not like:
            like = Oneup.query.filter_by(star_id=star.id, author_id=interestmodel.persona_id).all()


        content = star.text
        for planet in star.planets:
            if isinstance(planet, LinkPlanet):
                link = planet.url
                link_content = get_site_content(link)
                content += ' ' + link_content


        topics = topic_model.get_topics_text(content)

        if like:
            train_set_pos.append(topics)
        else:
            train_set_neg.append(topics)


    app.logger.info("Fitting persona %s"%interestmodel.persona_id)
    app.logger.info("Positive: %d    Negative: %d"%(len(train_set_pos),len(train_set_neg)))
    if len(train_set_pos) > 0:
        train_labels = [1 for x in range(len(train_set_pos))]
        train_set = train_set_pos

        if len(train_set_neg) > 0:
            train_set.extend(train_set_neg)
            train_labels.extend([0 for x in range(len(train_set_neg))])

        interestmodel.classifier.fit(train_set, train_labels)

    interestmodel.last_fit = datetime.now()



from operator import itemgetter
from web_ui import app


class PageManager(object):
    """ Holds all factory methods for Page creation.
    Those methods read in possible layouts, and create
    and a Page instance arrangeing content in the most
    beneficial way. Also see Page class below.
    """

    def __init__(self):
        """ Initializes screen size and  loads all layouts
        for the right context, using the context defined in
        the concrete subclass.
        """

        self.screen_size = (12.0, 8.0)
        self.all_layouts = app.config['LAYOUT_DEFINITIONS']

    def _add_static_section(self, page, section, layout):
        # if section contains only one cell it's not a list
        if not isinstance(layout[section][0], list):
            page.add_to_section(section, layout[section], None)
            return

        for cell in layout[section]:
            page.add_to_section(section, cell, None)

    def _get_layouts_for(self, context):
        """ Returns all layouts appropriate for context """

        return [layout for layout in self.all_layouts if
                context in layout['context']]

    def group_layout(self, stars):
        context = 'group_page'
        layouts = self._get_layouts_for(context)

        # currently no logic to choose among different layouts
        assert(len(layouts) == 1)
        best_layout = layouts[0]

        page = Page()

        # Add header to group page
        section = 'header'
        page.add_to_section(section, best_layout[section], None)

        # Add create_star form to page
        section = 'create_star_form'

        for cell in layouts[0][section]:
            page.add_to_section(section, cell, None)

        # Add the stars of the group to page
        section = 'stars'

        # Rank stars by score
        stars_ranked = sorted(stars, key=lambda s: s.hot(), reverse=True)

        for i, star_cell in enumerate(best_layout[section]):
            if i >= len(stars_ranked):
                break

            star = stars_ranked[i]
            page.add_to_section(section, star_cell, star)

        return page

    def create_star_layout(self):
        """Returns a page for creating stars."""

        # use layouts for create_star_page context
        context = 'create_star_page'
        layouts = self._get_layouts_for(context)

        # currently no logic to choose among different layouts
        assert(len(layouts) == 1)

        # page entry containing the form will be called:
        section = 'create_star_form'

        page = Page()

        for cell in layouts[0][section]:
            page.add_to_section(section, cell, None)

        return page

    def create_group_layout(self):
        """Returns a page for creating groups."""

        # use layouts for create_group_page context
        context = 'create_group_page'
        layouts = self._get_layouts_for(context)
        page = Page()

        # currently no logic to choose among different layouts
        assert(len(layouts) == 1)
        best_layout = layouts[0]

        # add layout for static fields
        self._add_static_section(page, 'header', best_layout)
        self._add_static_section(page, 'create_group_form', best_layout)

        return page

    def star_layout(self, stars):
        """Return the optimal layouted page for the given stars."""

        context = 'star_page'
        layouts = self._get_layouts_for(context)

        section = 'stars'

        # Rank stars by score
        stars_ranked = sorted(stars, key=lambda s: s.hot(), reverse=True)

        # Find best layout by filling each one with stars
        # and determining which one gives the best score
        layout_scores = dict()
        for layout in layouts:
            # print("\nLayout: {}".format(layout['name']))
            layout_scores[layout['name']] = 0

            for i, star_cell in enumerate(layout[section]):
                if i >= len(stars_ranked):
                    continue
                star = stars_ranked[i]

                cell_score = self._cell_score(star_cell)
                layout_scores[layout['name']] += star.hot() * cell_score
                # print("{}\t{}\t{}".format(star, star.hot() * cell_score, cell_score))
            # print("Score: {}".format(layout_scores[layout['name']]))

        # Select best layout
        selected_layouts = sorted(
            layout_scores.iteritems(),
            key=itemgetter(1),
            reverse=True)

        if len(selected_layouts) == 0:
            # TODO: Throw exception here, if no layout found the PM failed
            app.logger.error("No fitting layout found")
            return

        for layout in layouts:
            if layout['name'] == selected_layouts[0][0]:
                break

        # print("Chosen {}".format(layout))

        page = Page()
        for i, star_cell in enumerate(layout[section]):
            if i >= len(stars_ranked):
                break

            star = stars_ranked[i]
            page.add_to_section(section, star_cell, star)

        return page

    def _cell_score(self, cell):
        """ Return a score that describes how valuable a given cell on
        the screen is. Cell on the top left is 1.0, score diminishes to
        the right and bottom. Bigger cells get higher scores, cells off
        the screen get a 0.0 """

        import math

        # position score
        if cell[0] > self.screen_size[0] or cell[1] > self.screen_size[1]:
            pscore = 0.0
        else:
            score_x = 1.0 if cell[0] == 0 else 1.0 / (1.0 + (cell[0] / self.screen_size[0]))
            score_y = 1.0 if cell[1] == 0 else 1.0 / (1.0 + (cell[1] / self.screen_size[1]))
            pscore = (score_x + score_y) / 2.0

        # size score (sigmoid)
        area = cell[2] * cell[3]
        sscore = 1.0 / (1 + pow(math.exp(1), -0.1 * (area - 12.0)))
        return pscore * sscore


class Page(object):
    """Responsible for the layout of a page. Instances of Page
    hold variables for each dynamic section of the page. This
    variables hold a list with the entries of the section.
    An entry is a dict with the keys 'css_class' and 'content'.
    Example:
        star_page:
            + header:[
                {css_class:'col1 row1 w3 h1', content: None}
              ]
            + stars: [
                {css_class:'col1 row2 w3 h1', content: star1},
                {css_class:'col1 row3 w3 h1', content: star2}
              ]
    """

    def add_to_section(self, section, entry, content):
        """ Adds a new entry to the page section 'section'
        (and creates it if necessary). """

        section_entry = self._create_entry(entry, content)

        if not hasattr(self, section):
            setattr(self, section, [])

        attr = getattr(self, section)
        attr.append(section_entry)

    def _create_entry(self, cell, content):
        """ Creates a section of a page consisting of a dict
        containing the css_class and the content of the section.
        CSS class name format
         col   column at which the css container begins
         row   row at which it begins
         w     width of the container
         h     height of the container
        """
        css_class = "col{} row{} w{} h{}".format(
            cell[0],
            cell[1],
            cell[2],
            cell[3])

        return {'css_class': css_class, 'content': content}



import logging
import os
import sys
import argparse

from Crypto.Hash import SHA256
from flask import Flask, json
from flask.ext.sqlalchemy import SQLAlchemy
from flaskext import uploads
from humanize import naturaltime
from werkzeug.contrib.cache import SimpleCache


# Initialize Flask app
app = Flask('souma')

# Load config from default_config.py
app.config.from_object("web_ui.default_config")
app.jinja_env.filters['naturaltime'] = naturaltime

# Create application data folder
if not os.path.exists(app.config["USER_DATA"]):
    os.mkdir(app.config["USER_DATA"], 0700)

# Parse command line arguments
parser = argparse.ArgumentParser(description='Start Souma client')
parser.add_argument('--no_ui',
    default=False,
    action="store_true",
    help="skip starting the web ui server")

parser.add_argument('-v',
    '--verbose',
    default=False,
    action="store_true",
    help="gimme moar-logs")

parser.add_argument('-r',
    '--reset',
    default=False,
    action="store_true",
    help="reset database and secret key (irreversible!)")

parser.add_argument('-p',
    '--port',
    type=int,
    help='run synapse on this port')

parser.add_argument('-g',
    '--glia',
    default=app.config['LOGIN_SERVER'],
    help="glia server url")

args = parser.parse_args()
app.config['NO_UI'] = args.no_ui
app.config['LOGIN_SERVER'] = args.glia

if args.verbose is True:
    app.config["LOG_LEVEL"] = logging.DEBUG
    app.logger.debug("Verbose logs active")

if args.port is not None:
    app.config['LOCAL_PORT'] = args.port
    app.config['LOCAL_ADDRESS'] = "{}:{}".format(app.config['LOCAL_HOSTNAME'], args.port)
    app.config['SYNAPSE_PORT'] = args.port + 50
    app.config['DATABASE'] = os.path.join(app.config["USER_DATA"], 'souma_{}.db'.format(app.config["LOCAL_PORT"]))
    app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite:///" + app.config['DATABASE']
    app.config["SECRET_KEY_FILE"] = os.path.join(app.config["USER_DATA"], "secret_key_{}.dat".format(args.port))
    app.config["PASSWORD_HASH_FILE"] = os.path.join(app.config["USER_DATA"], "pw_hash_{}.dat".format(args.port))

if args.reset is True:
    for fileid in ["DATABASE", "SECRET_KEY_FILE", "PASSWORD_HASH_FILE"]:
        try:
            os.remove(app.config[fileid])
        except OSError:
            app.logger.warning("RESET: {} not found".format(fileid))
        else:
            app.logger.warning("RESET: {} deleted")


""" Load/set secret key """
try:
    with open(app.config["SECRET_KEY_FILE"], 'rb') as f:
        app.config['SECRET_KEY'] = f.read(24)
except IOError:
    # Create new secret key
    app.logger.debug("Creating new secret key")
    app.config['SECRET_KEY'] = os.urandom(24)
    with open(app.config["SECRET_KEY_FILE"], 'wb') as f:
        os.chmod(app.config["SECRET_KEY_FILE"], 0700)
        f.write(app.config['SECRET_KEY'])

# Generate ID used to identify this machine
app.config['SOUMA_ID'] = SHA256.new(app.config['SECRET_KEY'] + str(app.config['LOCAL_PORT'])).hexdigest()[:32]

if 'SOUMA_PASSWORD_HASH_{}'.format(app.config['LOCAL_PORT']) in os.environ:
    app.config['PASSWORD_HASH'] = os.environ['SOUMA_PASSWORD_HASH_{}'.format(app.config["LOCAL_PORT"])]
else:
    app.config['PASSWORD_HASH'] = None


# Load layout definitions
try:
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'layouts.json')) as f:
        app.config['LAYOUT_DEFINITIONS'] = json.load(f)
except IOError, e:
    logging.error("Failed loading layout definitions")
    app.config['LAYOUT_DEFINITIONS'] = dict()


# Setup SQLAlchemy database
db = SQLAlchemy(app)


# Setup attachment access
attachments = uploads.UploadSet('attachments', uploads.IMAGES)
uploads.configure_uploads(app, (attachments))


# Setup loggers
# Flask is configured to route logging events only to the console if it is in debug
# mode. This overrides this setting and enables a new logging handler which prints
# to the shell.
loggers = [app.logger, logging.getLogger('synapse'), logging.getLogger('e-synapse')]
console_handler = logging.StreamHandler(stream=sys.stdout)
console_handler.setFormatter(logging.Formatter(app.config['LOG_FORMAT']))

for l in loggers:
    del l.handlers[:]  # remove old handlers
    l.setLevel(logging.DEBUG)
    l.addHandler(console_handler)
    l.propagate = False  # setting this to true triggers the root logger


# Log loaded configuration info
app.logger.info(
    "\n".join(["{:=^80}".format(" SOUMA CONFIGURATION "),
              "{:>12}: {}".format("souma", app.config['SOUMA_ID'][:6]),
              "{:>12}: {}".format("web ui", "disabled" if app.config['NO_UI'] else app.config['LOCAL_ADDRESS']),
              "{:>12}: {}:{}".format(
                  "synapse",
                  app.config['LOCAL_HOSTNAME'],
                  app.config['SYNAPSE_PORT']),
              "{:>12}: {}".format("database", app.config['DATABASE']),
              "{:>12}: {}".format("glia server", app.config['LOGIN_SERVER'])]))


# Setup Cache
cache = SimpleCache()


def logged_in():
    """Check whether a user is logged in"""
    return cache.get('password') is not None

# Views need to be imported at the bottom to avoid circular import (see Flask docs)
import web_ui.views



import json
import datetime
import iso8601
import keyczar

from base64 import b64encode, b64decode
from hashlib import sha256
from keyczar.keys import AesKey, HmacKey

from nucleus import PersonaNotFoundError, InvalidSignatureError, UnauthorizedError, VesicleStateError
from nucleus.models import Persona
from web_ui import app, db

VESICLE_VERSION = "0.1"
DEFAULT_ENCODING = "{version}-{encoding}".format(version=VESICLE_VERSION, encoding="plain")
SYNAPSE_PORT = None
AES_BYTES = 256


class Vesicle(db.Model):
    """
    Container for peer messages

    see https://github.com/ciex/souma/wiki/Vesicle

    """
    __tablename__ = "vesicle"
    _default_send_attributes = ["message_type", "id", "payload", "enc"]

    id = db.Column(db.String(32), primary_key=True)
    message_type = db.Column(db.String(32))
    payload = db.Column(db.Text)
    signature = db.Column(db.Text)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    keycrypt = db.Column(db.Text(), default='{}')
    enc = db.Column(db.String(16))
    _send_attributes = db.Column(db.Text)
    handled = db.Column(db.Boolean)

    author_id = db.Column(db.String(32), db.ForeignKey('persona.id', use_alter=True, name="fk_author_id"))
    author = db.relationship('Persona', primaryjoin="Persona.id==Vesicle.author_id", post_update=True)

    _hashcode = None
    data = None

    def __init__(self, id, message_type, author=None, data=None, payload=None,
            signature=None, created=None, keycrypt='{}', enc=DEFAULT_ENCODING, handled=False):

        db.Model.__init__(self,
            id=id,
            message_type=message_type,
            author=author,
            data=data,
            payload=payload,
            signature=signature,
            created=created,
            keycrypt=keycrypt,
            enc=enc,
            handled=handled)

        self._send_attributes = json.dumps(Vesicle._default_send_attributes)

    def __str__(self):
        """Return string identifier

        Returns:
            String: Identifier
        """
        if hasattr(self, "author_id") and self.author_id is not None:
            if self.author is not None:
                author = self.author
            else:
                author = "<[{}]>".format(self.author_id[:6])
        else:
            author = "anon"

        try:
            if not self.decrypted():
                desc = "encrypted"
            elif self.message_type == "object_request":
                desc = "<{} [{}]>".format(self.data["type"], self.data["id"][:6])
            else:
                desc = "{}-<{} [{}]>".format(self.data["action"], self.data["object_type"], self.data["obj"]["id"])
        except KeyError:
            desc = "unknown"

        return "<Vesicle {mt}:{desc} by {author} [{id}]>".format(
            mt=self.message_type,
            desc=desc,
            author=author,
            id=self.id[:6])

    def _get_hashcode(self, payload=None):
        """Return the sha256 hashcode of unencrypted self.payload

        If the hashcode is cached, that will be returned. Otherwise, it will
        be computed if Vesicle is not encrypted and has a payload.

        Args:
            payload (String): (Optional) data to use instead of self.payload

        Returns:
            String: The hashcode

        Raises:
            VesicleStateError: If Vesicle is encrypted and hashcode is not cached
            ValueError: If no payload was provided by argument and self.payload is None
        """
        if self._hashcode is None:
            if self.encrypted():
                raise VesicleStateError("Error computing hashcode: {} is encrypted".format(self))

            if payload is None:
                if self.payload is None:
                    raise ValueError("Error computing hashcode: {} has no payload".format(self))
                payload = self.payload

            # AES keys can not be longer than 32 bytes
            self._hashcode = sha256(payload).hexdigest()[:32]

        return self._hashcode

    def encrypt(self, recipients):
        """
        Encrypt the vesicle's data field into the payload field using self.author and set the data field to None

        Args:
            recipients (list): A list of recipient Persona objects who will be added to the keycrypt
                author will always be added to the keycrypt!

        Raises:
            VesicleStateError: If this Vesicle is already encrypted
        """

        # Validate state
        if self.encrypted():
            raise VesicleStateError("Cannot encrypt already encrypted {}".format(self))

        # Retrieve a string representation of the message data
        if self.payload is not None:
            payload = self.payload
        else:
            if self.data == "" or self.data is None:
                raise ValueError("Cannot encrypt empty {} (`data` and `payload` fields are empty)".format(self))
            payload = json.dumps(self.data)

        # Generate an AES key with key=h

        # AES uses HMAC to authenticate its output. Usually this requires a key different from the
        # AES encryption key. Here I derive it from the AES key because authentication is already
        # provided by the separate RSA signature.
        # TODO: Ask someone whether this is a good idea
        self._hashcode = None
        h = self._get_hashcode(payload=payload)
        key = AesKey(h, HmacKey(h), AES_BYTES)

        # Encrypt payload using the AES key
        payload_encrypted = b64encode(key.Encrypt(payload))

        self.payload = payload_encrypted
        self.data = None
        self.signature = None  # must re-sign after encryption
        self.enc = "{version}-AES-{bytes}".format(version=self.enc.split("-")[0], bytes=AES_BYTES)

        self.add_send_attribute("author_id")
        self.add_send_attribute("keycrypt")

        self.add_recipients(recipients + [self.author], hashcode=h)

    def encrypted(self):
        """
        Return True if this Vesicle is encrypted

        Returns:
            Boolean: if encrypted
        """
        return self.payload is not None and self.enc.split("-")[1] != "plain"

    def decrypt(self, reader_persona=None):
        """
        Decrypt the vesicle's payload field into the data field.

        This method does not remove the ciphertext from the payload field, so that encrypted() still returns True.

        Args:
            reader_persona (Persona): Persona instance used to retrieve the hash key

        Returns:
            Persona: reader_persona used for decryption

        Raises:
            ValueError: If this Vesice is already plaintext
            UnauthorizedError: If no Persona was found for decrypting or reader_persona is not authorized
        """
        # Validate state
        if not self.encrypted():
            raise VesicleStateError("Cannot decrypt {}: Already plaintext.".format(self))

        keycrypt = json.loads(self.keycrypt)

        if reader_persona is None:
            for p in Persona.query.filter('sign_private != ""'):
                if p.id in keycrypt.keys():
                    reader_persona = p
                    continue

            if reader_persona is None:
                raise UnauthorizedError(
                    "Could not decrypt {}. No recipient found in owned personas.\nKeycrypt:{}".format(
                        self, keycrypt.keys()))
        else:
            if not reader_persona.id in keycrypt.keys():
                raise UnauthorizedError("No key found decrypting {} for {}".format(self, reader_persona))

        # Retrieve hashcode
        self._hashcode = reader_persona.decrypt(keycrypt[reader_persona.id])

        # Generate the AES key (see encrypt() above)
        key = AesKey(self._hashcode, HmacKey(self._hashcode), AES_BYTES)

        # Decrypt the data
        data = key.Decrypt(b64decode(self.payload))

        # Decode JSON
        self.data = json.loads(data)

        return reader_persona

    def decrypted(self):
        """
        Return True if this Vesice is decrypted

        Returns:
            Boolean: if decrypted
        """
        return self.data is not None

    def get_send_attributes(self):
        return json.loads(self._send_attributes)

    def set_send_attributes(self, send_attributes):
        self._send_attributes = json.dumps(send_attributes)

    def add_send_attribute(self, attr):
        send_attributes = self.get_send_attributes()
        if send_attributes.count(attr) == 0:
            send_attributes.append(attr)
        self.set_send_attributes(send_attributes)

    def remove_send_attribute(self, attr):
        send_attributes = self.get_send_attributes()
        send_attributes.remove(attr)
        self.set_send_attributes(send_attributes)

    def sign(self):
        """
        Sign a vesicle using vesicle.author's signing key

        Raises:
            KeyError: If no author is defined
        """
        if self.author is None:
            raise KeyError("Vesicle defines no author for signing")

        if not self.encrypted():
            raise VesicleStateError("Vesicle must be encrypted for signing")

        self.signature = self.author.sign(self.payload)

        self.add_send_attribute("signature")
        self.add_send_attribute("author_id")

    def signed(self):
        """
        Return True if vesicle has a signature and it is valid

        Returns:
            Boolean: If signed
        """

        if not hasattr(self, "signature"):
            return False

        if not self.author:
            raise PersonaNotFoundError(self.author_id)

        return self.author.verify(self.payload, self.signature)

    def add_recipients(self, recipients, hashcode=None):
        """
        Add Personas to the keycrypt

        Args:
            recipient (List): List of Persona instances to be added
            hashcode (Str): (Optional) hashcode to use (for example when adding
                recipients to an encrypted Vesicle)

        Raises:
            VesicleStateError: When trying to add recipients to a plaintext Vesicle
            UnauthorizedError: If no hashcode provided and self.author is not controlled
        """
        if not self.encrypted():
            raise VesicleStateError("Can not add recipients to plaintext vesicles")

        keycrypt = json.loads(self.keycrypt)

        if hashcode is None:
            if not self.author.controlled():
                raise UnauthorizedError("Must be Vesicle author to add recipients without known hashcode")
            hashcode = self.author.decrypt(keycrypt[self.author_id])

        for recipient in recipients:
            if recipient.id not in keycrypt.keys():
                key = recipient.encrypt(hashcode)
                keycrypt[recipient.id] = key
        self.keycrypt = json.dumps(keycrypt)

    def remove_recipient(self, recipient):
        """
        Remove a persona from the keycrypt

        Args:
            recipient (Persona): Persona to be removed from the keycrypt

        Raises:
            UnauthorizedError: If trying to remove author from keycrypt
        """
        if recipient.id == self.author_id:
            raise UnauthorizedError("Can't remove author from keycrypt")

        keycrypt = json.loads(self.keycrypt)
        del keycrypt[recipient.id]
        self.keycrypt = json.dumps(keycrypt)
        app.logger.info("Removed {} as a recipient of {}".format(recipient, self))

    def json(self):
        """
        Generate JSON representation of this Vesicle, including all attributes defined in self._send_attributes.

        Returns:
            String: JSON encoded Vesicle contents
        """
        # Temporarily encode data if this is a plaintext message
        if self.payload is None:
            plainenc = True
            self.payload = json.dumps(self.data)
        else:
            plainenc = False

        message = dict()
        for attr in self.get_send_attributes():
            message[attr] = getattr(self, attr)
        message["created"] = datetime.datetime.utcnow().isoformat()
        message["souma_id"] = app.config["SOUMA_ID"]
        r = json.dumps(message)

        if plainenc:
            self.payload = None
        return r

    @staticmethod
    def read(data):
        """
        Create a vesicle instance from its JSON representation. Checks signature validity.

        Args:
            data (String): JSON representation of a Vesicle

        Returns:
            Vesicle: The newly read Vesicle object

        Raises:
            ValueError: Unknown protocol version or malformed created timestamp
            KeyError: Missing key in vesicle JSON
            InvalidSignatureError: Does not match author_id's pubkey
            PersonaNotFoundError: Vesicle author not found
        """

        msg = json.loads(data)

        version, encoding = msg["enc"].split("-", 1)
        if version != VESICLE_VERSION:
            raise ValueError("Unknown protocol version: {} \nExpecting: {}".format(version, VESICLE_VERSION))

        try:
            vesicle = Vesicle(
                id=msg["id"],
                message_type=msg["message_type"],
                payload=msg["payload"],
                keycrypt=msg["keycrypt"] if "keycrypt" in msg else "{}",
                created=iso8601.parse_date(msg["created"]).replace(tzinfo=None),
                enc=msg["enc"])

            send_attributes = msg.keys()
            send_attributes.remove('souma_id')
            vesicle.send_attributes = send_attributes

            vesicle.author = Persona.query.get(msg["author_id"])
            if vesicle.author is None:
                raise PersonaNotFoundError(msg["author_id"])

            if "signature" in msg:
                vesicle.signature = msg["signature"]
        except KeyError, e:
            raise KeyError(e)
        except iso8601.ParseError, e:
            raise ValueError("Vesicle malformed: Error parsing date ({})".format(e))

        # Verify signature
        if vesicle.signature is not None and not vesicle.signed():
            raise InvalidSignatureError("Invalid signature on {}".format(vesicle))

        return vesicle



import datetime
import os

from flask import url_for
from hashlib import sha256
from keyczar.keys import RsaPrivateKey, RsaPublicKey
from web_ui import db
from web_ui.helpers import Serializable, epoch_seconds
from sqlalchemy import ForeignKey
from sqlalchemy.exc import OperationalError

#
# Setup follower relationship on Persona objects
#

t_contacts = db.Table(
    'contacts',
    db.Column('left_id', db.String(32), db.ForeignKey('persona.id')),
    db.Column('right_id', db.String(32), db.ForeignKey('persona.id'))
)


class Persona(Serializable, db.Model):
    """A Persona represents a user profile"""

    __tablename__ = "persona"
    id = db.Column(db.String(32), primary_key=True)
    username = db.Column(db.String(80))
    email = db.Column(db.String(120))
    contacts = db.relationship(
        'Persona',
        secondary='contacts',
        primaryjoin='contacts.c.left_id==persona.c.id',
        secondaryjoin='contacts.c.right_id==persona.c.id')
    crypt_private = db.Column(db.Text)
    crypt_public = db.Column(db.Text)
    sign_private = db.Column(db.Text)
    sign_public = db.Column(db.Text)
    modified = db.Column(db.DateTime, default=datetime.datetime.now(), onupdate=datetime.datetime.now())
    soma_id = db.Column(db.String(32), db.ForeignKey('starmap.id'))
    soma = db.relationship('Starmap', backref="personas", primaryjoin='starmap.c.id==persona.c.soma_id')

    def __init__(self, id, username, email=None, sign_private=None, sign_public=None,
                 crypt_private=None, crypt_public=None):

        self.id = id
        self.username = username
        self.email = email
        self.sign_private = sign_private
        self.sign_public = sign_public
        self.crypt_private = crypt_private
        self.crypt_public = crypt_public

    def __repr__(self):
        return "<{} [{}]>".format(str(self.username), self.id[:6])

    def get_email_hash(self):
        """Return sha256 hash of this user's email address"""
        return sha256(self.email).hexdigest()

    def get_absolute_url(self):
        return url_for('persona', id=self.id)

    def generate_keys(self, password):
        """ Generate new RSA keypairs for signing and encrypting. Commit to DB afterwards! """

        # TODO: Store keys encrypted
        rsa1 = RsaPrivateKey.Generate()
        self.sign_private = str(rsa1)
        self.sign_public = str(rsa1.public_key)

        rsa2 = RsaPrivateKey.Generate()
        self.crypt_private = str(rsa2)
        self.crypt_public = str(rsa2.public_key)

    def encrypt(self, data):
        """ Encrypt data using RSA """

        key_public = RsaPublicKey.Read(self.crypt_public)
        return key_public.Encrypt(data)

    def decrypt(self, cypher):
        """ Decrypt cyphertext using RSA """

        key_private = RsaPrivateKey.Read(self.crypt_private)
        return key_private.Decrypt(cypher)

    def sign(self, data):
        """ Sign data using RSA """
        from base64 import b64encode

        key_private = RsaPrivateKey.Read(self.sign_private)
        signature = key_private.Sign(data)
        return b64encode(signature)

    def verify(self, data, signature_b64):
        """ Verify a signature using RSA """
        from base64 import b64decode

        signature = b64decode(signature_b64)
        key_public = RsaPublicKey.Read(self.sign_public)
        return key_public.Verify(data, signature)


class Star(Serializable, db.Model):
    """A Star represents a post"""

    __tablename__ = "star"
    id = db.Column(db.String(32), primary_key=True)
    text = db.Column(db.Text)
    created = db.Column(db.DateTime, default=datetime.datetime.now())
    modified = db.Column(db.DateTime, default=datetime.datetime.now(), onupdate=datetime.datetime.now())

    planets = db.relationship(
        'Planet',
        secondary='satellites',
        backref=db.backref('starmap'),
        primaryjoin="satellites.c.star_id==star.c.id",
        secondaryjoin="satellites.c.planet_id==planet.c.id")

    creator = db.relationship(
        'Persona',
        backref=db.backref('starmap'),
        primaryjoin="Persona.id==Star.creator_id")

    creator_id = db.Column(db.String(32), db.ForeignKey('persona.id'))

    def __init__(self, id, text, creator):
        self.id = id
        # TODO: Attach multiple items as 'planets'
        self.text = text

        if not isinstance(creator, Persona):
            self.creator_id = creator
        else:
            self.creator_id = creator.id

    def __repr__(self):
        ascii_text = self.text.encode('utf-8')
        return "<Star {}: {}>".format(
            self.creator_id[:6],
            (ascii_text[:24] if len(ascii_text) <= 24 else ascii_text[:22] + ".."))

    def get_absolute_url(self):
        return url_for('star', id=self.id)

    def hot(self):
        """i reddit"""
        from math import log
        # Uncomment to assign a score with analytics.score
        #s = score(self)
        s = 1.0
        order = log(max(abs(s), 1), 10)
        sign = 1 if s > 0 else -1 if s < 0 else 0
        return round(order + sign * epoch_seconds(self.created) / 45000, 7)


t_satellites = db.Table(
    'satellites',
    db.Column('star_id', db.String(32), db.ForeignKey('star.id')),
    db.Column('planet_id', db.String(32), db.ForeignKey('planet.id'))
)


class Planet(Serializable, db.Model):
    """A Planet represents an attachment"""

    __tablename__ = 'planet'
    id = db.Column(db.String(32), primary_key=True)
    title = db.Column(db.Text)
    kind = db.Column(db.String(32))
    created = db.Column(db.DateTime, default=datetime.datetime.now())
    modified = db.Column(db.DateTime, default=datetime.datetime.now(), onupdate=datetime.datetime.now())
    source = db.Column(db.String(128))

    __mapper_args__ = {
        'polymorphic_identity': 'planet',
        'polymorphic_on': kind
    }


class PicturePlanet(Planet):
    """A Picture attachment"""

    id = db.Column(db.String(32), ForeignKey('planet.id'), primary_key=True)
    filename = db.Column(db.Text)

    __mapper_args__ = {
        'polymorphic_identity': 'picture'
    }


class LinkPlanet(Planet):
    """A URL attachment"""

    id = db.Column(db.String(32), ForeignKey('planet.id'), primary_key=True)
    url = db.Column(db.Text)

    __mapper_args__ = {
        'polymorphic_identity': 'link'
    }


class Notification(db.Model):
    """Represents a stored notification to the user"""
    id = db.Column(db.String(32), primary_key=True)
    kind = db.Column(db.String(32))
    created = db.Column(db.DateTime, default=datetime.datetime.now())
    to_persona_id = db.Column(db.String(32))

    def __init__(self, kind, to_persona_id):
        from uuid import uuid4
        self.id = uuid4().hex
        self.kind = kind
        self.to_persona_id = to_persona_id



import appdirs
import logging
import os

# ANY CHANGES MADE TO OPTIONS DEPENDING ON LOCAL_PORT NEED TO BE UPDATED IN
# web_ui/__init__.py


def packaged_app():
    """Return True if currently running in a packaged context"""
    import sys
    if getattr(sys, 'frozen', None) == 'macosx_app':
        return True

USER_DATA = appdirs.user_data_dir("souma", "souma", roaming=True)

#
# --------------------- FLASK OPTIONS ---------------------
#

LOCAL_PORT = 5000
LOCAL_HOSTNAME = 'app.souma'
LOCAL_ADDRESS = "{}:{}".format(LOCAL_HOSTNAME, LOCAL_PORT)

DEBUG = False
USE_DEBUG_SERVER = False

SECRET_KEY_FILE = os.path.join(USER_DATA, "secret_key_{}.dat".format(LOCAL_PORT))
PASSWORD_HASH_FILE = os.path.join(USER_DATA, "pw_hash_{}.dat".format(LOCAL_PORT))

# Uncomment to log DB statements
# SQLALCHEMY_ECHO = True

DATABASE = os.path.join(USER_DATA, 'souma_{}.db'.format(LOCAL_PORT))
SQLALCHEMY_DATABASE_URI = "sqlite:///" + DATABASE

TOPIC_MODEL = os.path.join(USER_DATA, 'enwiki_lda.model')
TOPIC_MODEL_IDS = os.path.join(USER_DATA, 'enwiki__wordids.txt')
TOPIC_MODEL_UPDATE = "http://dl.dropboxusercontent.com/u/46877/topic_model/enwiki_lda.model"
TOPIC_MODEL_IDS_UPDATE = "http://dl.dropboxusercontent.com/u/46877/topic_model/enwiki__wordids.txt"

# uploads are placed in the UPLOADS_DEFAULT_DEST/'attachments' subfolder by flask-uploads
# this is configured in web_ui/__init__.py
UPLOADS_DEFAULT_DEST = os.path.join(os.path.abspath(os.path.dirname(__file__)), "static")
ALLOWED_EXTENSIONS = set(['txt', 'pdf', 'png', 'jpg', 'jpeg', 'gif'])
SEND_FILE_MAX_AGE_DEFAULT = 1

LOG_LEVEL = logging.INFO
LOG_FORMAT = (
    '%(name)s :: %(module)s [%(pathname)s:%(lineno)d]\n' +
    '%(message)s\n')

#
# --------------------- SYNAPSE OPTIONS -------------------
#

SYNAPSE_PORT = LOCAL_PORT + 2000

LOGIN_SERVER = "app.souma:24500"
# Uncomment this to use Heroku server
LOGIN_SERVER = "glia.herokuapp.com"

# Setting this to True will automatically upload all vesicles to Myelin, and
# enable periodic polling of the Myelin for new Vesicles sent to one of the
# Personas controlled by this Souma
ENABLE_MYELIN = True

# The interval in seconds at which the Myelin will be polled for new Vesicles
MYELIN_POLLING_INTERVAL = 10



from flask.ext.wtf import Form
from wtforms.fields import TextField, TextAreaField, SelectField, FileField, HiddenField
from wtforms.fields.html5 import URLField
from wtforms.validators import DataRequired, Email


class Create_persona_form(Form):
    """ Generate form for creating a persona """
    name = TextField('Name', validators=[DataRequired(), ])
    email = TextField('Email (optional)', validators=[Email(), ])


class Create_star_form(Form):
    """ Generate form for creating a star """
    # Choices of the author field need to be set before displaying the form
    # TODO: Validate author selection
    author = SelectField('Author', validators=[DataRequired(), ])
    text = TextAreaField('Content', validators=[DataRequired(), ])
    picture = FileField('Picture')
    #link = URLField('Link', validators=[url()])
    link = URLField('Link')
    context = HiddenField('Context', validators=[])


class Create_group_form(Form):
    """ Generate form for creating a group """

    author = HiddenField('Author', validators=[DataRequired(), ])
    groupname = TextField('Group name', validators=[DataRequired(), ])
    description = TextAreaField('Description', validators=[DataRequired(), ])


class FindPeopleForm(Form):
    email = TextField('Email-Address', validators=[
        DataRequired(), Email()])


class AddContactForm(Form):
    recipient_id = TextField('Persona ID', validators=[DataRequired()])
    author_id = SelectField('Contact as', validators=[DataRequired()])



#!/usr/bin/python

import requests
import sys
import webbrowser

from web_ui import app, db

from gevent import Greenlet
from gevent.wsgi import WSGIServer
from gevent.event import Event
from sqlalchemy.exc import OperationalError
from uuid import uuid4

from nucleus.models import Souma, Starmap
from synapse import Synapse

from astrolab.helpers import repeated_func_schedule
from astrolab.interestmodel import update


def setup_astrolab():
    """Download topic model and schedule model updates"""
    model_filename = app.config["TOPIC_MODEL"]
    word_ids_filename = app.config["TOPIC_MODEL_IDS"]

    model_url = app.config["TOPIC_MODEL_UPDATE"]
    word_ids_url = app.config["TOPIC_MODEL_IDS_UPDATE"]

    try:
        with open(word_ids_filename):
            app.logger.debug("Model word ids found")
    except IOError:
        app.logger.info("Now downloading model data ids")
        r = requests.get(word_ids_url, stream=True)
        with open(word_ids_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    f.flush()
        app.logger.info("Model data ids downloaded")

    try:
        with open(model_filename):
            app.logger.debug("Model data found")
    except IOError:
        app.logger.info("Downloading model data")
        r = requests.get(model_url, stream=True)
        with open(model_filename, 'wb') as f:
            i = 0
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    i += 1
                    if i % (1024 * 5) == 0:
                        app.logger.info("Downloaded {} MB / 323 MB".format(i / 1024))
        app.logger.info("Model data downloaded")

    repeated_func_schedule(60 * 60, update)

""" patch gevent for py2app """
if getattr(sys, 'frozen', None) == 'macosx_app':
        import imp
        import httplib

        original_load_module = imp.load_module
        original_find_module = imp.find_module

        def custom_load_module(name, file, pathname, description):
            if name == '__httplib__':
                return httplib
            return original_load_module(name, file, pathname, description)

        def custom_find_module(name, path=None):
            if name == 'httplib':
                return (None, None, None)
            return original_find_module(name, path)

        imp.load_module = custom_load_module
        imp.find_module = custom_find_module

        # Verify that the patch is working properly (you can remove these lines safely)
        __httplib__ = imp.load_module('__httplib__', *imp.find_module('httplib'))
        assert __httplib__ is httplib

""" Initialize database """
try:
    local_souma = Souma.query.get(app.config["SOUMA_ID"])
except OperationalError:
    local_souma = None

if local_souma is None:
    app.logger.info("Setting up database")
    db.create_all()

    app.logger.info("Setting up Nucleus for <Souma [{}]>".format(app.config['SOUMA_ID'][:6]))
    local_souma = Souma(id=app.config['SOUMA_ID'])
    local_souma.generate_keys()
    local_souma.starmap = Starmap(id=uuid4().hex, kind="index")

    db.session.add(local_souma)
    db.session.commit()

""" Start app """
if app.config['USE_DEBUG_SERVER']:
    # flask development server
    app.run(app.config['LOCAL_HOSTNAME'], app.config['LOCAL_PORT'])
else:
    shutdown = Event()

    # Synapse
    app.logger.info("Starting Synapses")
    synapse = Synapse()

    # Setup Astrolab
    Greenlet.spawn(setup_astrolab)

    # Web UI
    if not app.config['NO_UI']:
        app.logger.info("Starting Web-UI")
        local_server = WSGIServer(('', app.config['LOCAL_PORT']), app)
        local_server.start()
        webbrowser.open("http://{}".format(app.config["LOCAL_ADDRESS"]))

    shutdown.wait()



import os
import datetime

from flask import abort, flash, redirect, render_template, request, session, url_for, jsonify as json_response
from hashlib import sha256

from web_ui import app, cache, db, logged_in, attachments
from web_ui import pagemanager
from web_ui.forms import *
from web_ui.helpers import get_active_persona
from nucleus import notification_signals, PersonaNotFoundError
from nucleus.models import Persona, Star, Planet, PicturePlanet, LinkPlanet, Group, Starmap

# Create blinker signal namespace
local_model_changed = notification_signals.signal('local-model-changed')
contact_request_sent = notification_signals.signal('contact-request-sent')
new_contact = notification_signals.signal('new-contact')
group_created = notification_signals.signal('group-created')

pagemanager = pagemanager.PageManager()

@app.context_processor
def persona_context():
    """Makes controlled_personas available in templates"""
    return dict(
        controlled_personas=Persona.list_controlled(),
        active_persona=Persona.query.get(get_active_persona())
    )


@app.before_request
def before_request():
    """Preprocess requests"""

    session['active_persona'] = get_active_persona()

    def pass_thru(path):
        """Return True if no auth required for path"""
        allowed_paths = [
            '/setup',
            '/login'
        ]
        return path in allowed_paths or request.path[1:7] == 'static'

    if not pass_thru(request.path) and not logged_in():
        app.logger.info("Not logged in: Redirecting to Login")
        return redirect(url_for('login', _external=True))


@app.teardown_request
def teardown_request(exception):
    """Things to do after a request"""
    pass


@app.route('/login', methods=['GET', 'POST'])
def login():
    """Display a login form and create a session if the correct pw is submitted. Redirect to /setup if no pw hash."""
    from Crypto.Protocol.KDF import PBKDF2

    try:
        with open(app.config["PASSWORD_HASH_FILE"], "r") as f:
            password_hash = f.read()
    except IOError:
        app.logger.info("No password hash found: Redirecting to Setup")
        return redirect(url_for('setup', _external=True))

    error = None
    if request.method == 'POST':
        # TODO: Is this a good idea?
        salt = app.config['SECRET_KEY']
        pw_submitted = PBKDF2(request.form['password'], salt)

        if sha256(pw_submitted).hexdigest() != password_hash:
            error = 'Invalid password'
        else:
            cache.set('password', pw_submitted, 3600)
            flash('You are now logged in')
            return redirect(url_for('universe'))
    return render_template('login.html', error=error)


@app.route('/logout')
def logout():
    cache.set('password', None)
    flash('You were logged out')
    return redirect(url_for('login'))


@app.route('/setup', methods=['GET', 'POST'])
def setup():
    from Crypto.Protocol.KDF import PBKDF2
    from hashlib import sha256

    error = None
    if request.method == 'POST':
        logged_in()
        if request.form['password'] is None:
            error = 'Please enter a password'
        else:
            salt = app.config['SECRET_KEY']
            password = PBKDF2(request.form['password'], salt)
            password_hash = sha256(password).hexdigest()

            with open(app.config["PASSWORD_HASH_FILE"], "w") as f:
                f.write(password_hash)

            cache.set('password', password, 3600)
            return redirect(url_for('universe'))
    return render_template('setup.html', error=error)


@app.route('/p/<id>/')
def persona(id):
    """ Render home view of a persona """

    persona = Persona.query.filter_by(id=id).first_or_404()
    if hasattr(persona, "profile"):
        stars = persona.profile.index
    else:
        stars = []

    # TODO: Use new layout system
    vizier = Vizier([
        [1, 5, 6, 2],
        [1, 1, 6, 4],
        [7, 1, 2, 2],
        [7, 3, 2, 2],
        [7, 5, 2, 2]])

    return render_template(
        'persona.html',
        layout="persona",
        vizier=vizier,
        stars=stars,
        persona=persona)


@app.route('/p/create', methods=['GET', 'POST'])
def create_persona():
    """ Render page for creating new persona """
    from uuid import uuid4

    form = Create_persona_form()
    if form.validate_on_submit():
        # This is a unique ID which identifies the persona across all contexts
        uuid = uuid4().hex
        created_dt = datetime.datetime.utcnow()

        # Save persona to DB
        p = Persona(
            id=uuid,
            username=request.form['name'],
            email=request.form['email'],
            modified=created_dt,
        )

        # Create keypairs
        p.generate_keys(cache.get('password'))

        # TODO: Error message when user already exists
        db.session.add(p)
        db.session.commit()

        p.profile = Starmap(
            id=uuid4().hex,
            author=p,
            kind="persona_profile",
            modified=created_dt
        )
        db.session.add(p.profile)

        p.index = Starmap(
            id=uuid4().hex,
            author=p,
            kind="index",
            modified=created_dt
        )
        db.session.add(p.index)
        db.session.commit()

        local_model_changed.send(create_persona, message={
            "author_id": p.id,
            "action": "insert",
            "object_id": p.id,
            "object_type": "Persona",
        })

        local_model_changed.send(create_persona, message={
            "author_id": p.id,
            "action": "insert",
            "object_id": p.profile.id,
            "object_type": "Starmap",
        })

        local_model_changed.send(create_persona, message={
            "author_id": p.id,
            "action": "insert",
            "object_id": p.index.id,
            "object_type": "Starmap",
        })

        # Activate new Persona
        session["active_persona"] = p.id

        flash("New persona {} created!".format(p.username))
        return redirect(url_for('persona', id=uuid))

    return render_template(
        'create_persona.html',
        form=form,
        next=url_for('create_persona'))


@app.route('/p/<id>/activate', methods=['GET'])
def activate_persona(id):
    """ Activate a persona """
    p = Persona.query.get(id)
    if not p:
        app.logger.error("Tried to activate a nonexistent persona")
        abort(404)
    if p.sign_private == "":
        app.logger.error("Tried to activate foreign persona")
        flash("That is not you!")
    else:
        app.logger.info("Activated {}".format(p))
        session['active_persona'] = id
    return redirect(url_for('universe'))


@app.route('/s/create', methods=['GET', 'POST'])
def create_star():
    from uuid import uuid4
    """ Create a new star """

    form = Create_star_form(default_author=get_active_persona())
    form.author.choices = [(p.id, p.username) for p in Persona.list_controlled()]
    form.author.data = get_active_persona()

    # Default is posting to author profile
    if form.context.data is None:
        form.context.data = Persona.query.get(form.author.data).profile.id

    if form.validate_on_submit():
        uuid = uuid4().hex

        author = Persona.query.get(request.form["author"])
        if not author.controlled():
            app.logger.error("Can't create Star with foreign Persona {}".format(author))
            flash("Can't create Star with foreign Persona {}".format(author))
            return redirect(url_for('create_star')), 401

        new_star_created = datetime.datetime.utcnow()
        new_star = Star(
            id=uuid,
            text=request.form['text'],
            author=author,
            created=new_star_created,
            modified=new_star_created
        )
        db.session.add(new_star)
        db.session.commit()

        flash('New star created!')
        app.logger.info('Created new {}'.format(new_star))
        model_change_messages = list()

        if 'picture' in request.files and request.files['picture'].filename != "":
            # compute hash
            picture_hash = sha256(request.files['picture'].stream.read()).hexdigest()
            request.files['picture'].stream.seek(0)

            # create or get planet
            planet = Planet.query.filter_by(id=picture_hash[:32]).first()
            if not planet:
                app.logger.info("Creating new planet for submitted file")
                filename = attachments.save(request.files['picture'],
                    folder=picture_hash[:2], name=picture_hash[2:] + ".")
                planet = PicturePlanet(
                    id=picture_hash[:32],
                    filename=os.path.join(attachments.name, filename))
                db.session.add(planet)

            # attach to star
            new_star.planets.append(planet)

            # commit
            db.session.add(new_star)
            db.session.commit()
            app.logger.info("Attached {} to new {}".format(planet, new_star))

        if 'link' in request.form and request.form['link'] != "":
            link_hash = sha256(request.form['link']).hexdigest()[:32]
            planet = Planet.query.filter_by(id=link_hash).first()
            if not planet:
                app.logger.info("Storing new Link")
                planet = LinkPlanet(
                    id=link_hash,
                    url=request.form['link'])
                db.session.add(planet)

            new_star.planets.append(planet)
            db.session.add(new_star)
            db.session.commit()
            app.logger.info("Attached {} to new {}".format(planet, new_star))

        model_change_messages.append({
            "author_id": new_star.author.id,
            "action": "insert",
            "object_id": new_star.id,
            "object_type": "Star",
        })

        # Add new Star to a Starmap depending on context
        starmap = Starmap.query.get(request.form["context"])
        if starmap is not None:
            starmap.index.append(new_star)
            starmap.modified = new_star_created
            db.session.add(starmap)

            model_change_messages.append({
                "author_id": new_star.author.id,
                "action": "update",
                "object_id": starmap.id,
                "object_type": "Starmap"
            })

        db.session.commit()

        # db.session.expunge_all()  # Remove everything from session before sending signal
        for m in model_change_messages:
            local_model_changed.send(create_star, message=m)

        # if new star belongs to a group, show group page
        if starmap is None:
            return redirect(url_for('star', id=uuid))
        else:
            return redirect(starmap.get_absolute_url())

    page = pagemanager.create_star_layout()

    return render_template('create_star.html',
                           form=form,
                           page=page)


@app.route('/s/<id>/delete', methods=["GET"])
def delete_star(id):
    """ Delete a star """
    # TODO: Should only accept POST instead of GET

    # Load instance and author persona
    s = Star.query.get(id)
    if s is None:
        abort(404)

    if not s.author.controlled():
        flash("You are not allowed to delete {}'s Stars".format(s.author))
        app.logger.error("Tried to delete one of {}'s Stars".format(s.author))
        return redirect(request.referrer)
    else:
        s.set_state(-2)
        try:
            db.session.add(s)
            db.session.commit()
        except:
            app.logger.info("Error deleting {}".format(s))
            db.session.rollback()
        else:
            message_delete = {
                "author_id": s.author.id,
                "action": "delete",
                "object_id": s.id,
                "object_type": "Star",
            }

            local_model_changed.send(delete_star, message=message_delete)

            app.logger.info("Deleted star {}".format(id))
        return redirect(url_for('debug'))


@app.route('/')
def universe():
    """ Render the landing page """
    # return only stars that are not in a group context
    stars = Star.query.filter(Star.state >= 0).all()
    page = pagemanager.star_layout(stars)

    if len(persona_context()['controlled_personas'].all()) == 0:
        return redirect(url_for('create_persona'))

    if len(stars) == 0:
        return redirect(url_for('create_star'))

    return render_template('universe.html', page=page)


@app.route('/s/<id>/', methods=['GET'])
def star(id):
    """ Display a single star """
    star = Star.query.filter(Star.id == id, Star.state >= 0).first_or_404()
    author = Persona.query.filter_by(id=id)

    return render_template('star.html', layout="star", star=star, author=author)


@app.route('/s/<star_id>/1up', methods=['POST'])
def oneup(star_id):
    """
    Issue a 1up to a Star using the currently activated Persona

    Args:
        star_id (string): ID of the Star
    """
    star = Star.query.get_or_404(star_id)
    try:
        oneup = star.toggle_oneup()
    except PersonaNotFoundError:
        error_message = "Please activate a Persona for upvoting"
        oneup = None

    resp = dict()
    if oneup is None:
        resp = {
            "meta": {
                "oneup_count": star.oneup_count(),
                "error_message": error_message
            }
        }
    else:
        resp = {
            "meta": {
                "oneup_count": star.oneup_count(),
            },
            "oneups": [{
                "id": oneup.id,
                "author": oneup.author.id,
                "state_value": oneup.state,
                "state_name": oneup.get_state()
            }]
        }
    return json_response(resp)


@app.route('/debug/')
def debug():
    """ Display raw data """
    stars = Star.query.all()
    personas = Persona.query.all()
    planets = Planet.query.all()
    groups = Group.query.all()
    starmaps = Starmap.query.all()

    return render_template(
        'debug.html',
        stars=stars,
        personas=personas,
        planets=planets,
        groups=groups,
        starmaps=starmaps
    )


@app.route('/find-people', methods=['GET', 'POST'])
def find_people():
    """Search for and follow people"""
    from synapse.electrical import ElectricalSynapse

    form = FindPeopleForm(request.form)
    error = found = None
    found_processed = list()
    found_new = list()

    if request.method == 'POST' and form.validate():
        # Compile message
        address = request.form['email']

        # Create a temporary electrical synapse to make a synchronous glia request
        electrical = ElectricalSynapse(None)
        resp, errors = electrical.find_persona(address)

        # TODO: This should flash an error message. It doesn't.
        if errors:
            flash("Server error: {}".format(str(errors)))

        elif resp and resp['personas']:
            found = resp['personas']
            active_persona = Persona.query.get(get_active_persona())

            for p in found:
                p_local = Persona.query.get(p['id'])
                if p_local is None:
                    import iso8601
                    app.logger.info("Storing new Persona {}".format(p['id']))

                    modified = iso8601.parse_date(p["modified"]).replace(tzinfo=None)

                    p_new = Persona(
                        id=p['id'],
                        username=p['username'],
                        modified=modified,
                        email=address,
                        crypt_public=p['crypt_public'],
                        sign_public=p['sign_public']
                    )
                    p_new._stub = True
                    found_new.append(p_new)
                    found_processed.append({
                        "id": p["id"],
                        "username": p["username"],
                        "incoming": None,
                        "outgoing": None
                    })
                else:
                    found_processed.append({
                        "id": p["id"],
                        "username": p["username"],
                        "incoming": (active_persona in p_local.contacts),
                        "outgoing": (p_local in active_persona.contacts)
                    })

            try:
                for p in found_new:
                    db.session.add(p)
                db.session.commit()
            except:
                db.session.rollback()
        else:
            error = "No record for {}. Check the spelling!".format(address)

    return render_template('find_people.html', form=form, found=found_processed, error=error)


@app.route('/p/<persona_id>/add_contact', methods=['GET', "POST"])
def add_contact(persona_id):
    """Add a persona to the current persona's address book"""
    form = AddContactForm(request.form)
    persona = Persona.query.get(persona_id)
    author = Persona.query.get(get_active_persona())

    if request.method == 'POST' and persona is not None:
        author.contacts.append(persona)
        db.session.add(author)
        db.session.commit()

        app.logger.info("Now sharing {}'s posts with {}".format(author, persona))

        local_model_changed.send(add_contact, message={
            "author_id": author.id,
            "action": "update",
            "object_id": author.id,
            "object_type": "Persona"
        })

        new_contact.send(add_contact, message={
            'new_contact_id': persona.id,
            'author_id': author.id
        })

        flash("Now sharing {}'s posts with {}".format(author, persona))
        return redirect(url_for('persona', id=persona.id))

    return render_template('add_contact.html', form=form, persona=persona)


@app.route('/g/<id>/', methods=['GET'])
def group(id):
    """ Render home view of a group """

    group = Group.query.filter_by(id=id).first_or_404()

    form = Create_star_form(default_author=get_active_persona())
    form.author.choices = [(p.id, p.username) for p in Persona.list_controlled()]

    # Fill in group-id to be used in star creation
    form.context.data = group.profile.id

    # create layouted page for group
    page = pagemanager.group_layout(group.profile.index.filter(Star.state >= 0))

    return render_template(
        'group.html',
        group=group,
        page=page,
        form=form)


@app.route('/g/create', methods=['GET', 'POST'])
def create_group():
    """ Render page for creating new group """

    from uuid import uuid4

    form = Create_group_form()
    form.author.choices = [(p.id, p.username) for p in Persona.list_controlled()]
    form.author.data = get_active_persona()

    if form.validate_on_submit():
        # create ID to identify the group across all contexts
        uuid = uuid4().hex
        created_dt = datetime.datetime.utcnow()

        author = Persona.query.get(request.form["author"])
        if not author.controlled():
            app.logger.error("Can't create Group with foreign Persona {}".format(author))
            flash("Can't create Group with foreign Persona {}".format(author))
            return redirect(url_for('create_group')), 401

        # Create group and add to DB
        g = Group(
            id=uuid,
            author=author,
            modified=created_dt,
            groupname=request.form['groupname'],
            description=request.form['description']
        )

        db.session.add(g)
        db.session.commit()

        index = Starmap(
            id=uuid4().hex,
            author=author,
            kind="group_profile",
            modified=created_dt
        )
        g.profile = index
        db.session.add(index)
        db.session.commit()

        flash("New group {} created!".format(g.groupname))
        app.logger.info("Created {} with {}".format(g, g.profile))

        local_model_changed.send(create_group, message={
            "author_id": author.id,
            "action": "insert",
            "object_id": g.id,
            "object_type": "Group",
        })

        local_model_changed.send(create_group, message={
            "author_id": author.id,
            "action": "insert",
            "object_id": g.profile.id,
            "object_type": "Starmap",
        })

        return redirect(url_for('group', id=g.id))

    page = pagemanager.create_group_layout()
    return render_template(
        'create_group.html',
        form=form,
        page=page,
        next=url_for('create_group'))


@app.route('/g/', methods=['GET'])
def groups():
    groups = Group.query.all()

    return render_template(
        'groups.html',
        groups=groups
    )


class Vizier():
    """Old layout system. Use Pagemanager instead"""
    def __init__(self, layout):
        from collections import defaultdict

        cells = defaultdict(list)
        for e in layout:
            x_pos = e[0]
            y_pos = e[1]
            x_size = e[2]
            y_size = e[3]

            for col in xrange(x_pos, x_pos + x_size):
                for row in xrange(y_pos, y_pos + y_size):
                    if col in cells and row in cells[col]:
                        app.logger.warning("Double binding of cell ({x},{y})".format(x=col, y=row))
                    else:
                        cells[col].append(row)

        self.layout = layout
        self.index = 0

    def get_cell(self):
        """ Return the next free cell's class name """
        if len(self.layout) <= self.index:
            app.logger.warning("Not enough layout cells provided for content.")
            return "hidden"

        class_name = "col{c} row{r} w{width} h{height}".format(
            c=self.layout[self.index][0],
            r=self.layout[self.index][1],
            width=self.layout[self.index][2],
            height=self.layout[self.index][3])

        self.index += 1
        return class_name



import datetime
import json
import logging
import requests
import iso8601

from base64 import b64encode
from Crypto import Random
from dateutil.parser import parse as dateutil_parse
from gevent import Greenlet
from hashlib import sha256
from humanize import naturaltime
from operator import itemgetter

from nucleus import notification_signals, ERROR, create_session
from nucleus.models import Persona, Souma
from web_ui import app

API_VERSION = 0
API_VERSION_LONG = 0.1


class GliaAuth(requests.auth.AuthBase):
    """Attaches HTTP Glia Authentication to the given Request object."""
    def __init__(self, souma, payload=None):
        self.souma = souma
        self.payload = payload if payload is not None else str()
        self.rng = Random.new()

    def __call__(self, r):
        # modify and return the request
        rand = self.rng.read(16)

        # app.logger.debug("Authenticating {}\nID: {}\nRand: {}\nPath: {}\nPayload: {}".format(
        #    r, str(self.souma.id), rand, str(r.url), self.payload))

        r.headers['Glia-Souma'] = self.souma.id
        r.headers['Glia-Rand'] = b64encode(rand)
        r.headers['Glia-Auth'] = self.souma.sign("".join([
            str(self.souma.id),
            rand,
            str(r.url),
            self.payload
        ]))
        return r


class ElectricalSynapse(object):
    """
    Handle connection to HTTP endpoints

    Parameters:
        parent (Synapse) A Synapse object to which received Vesicles will be passed
        host (String) The IP address of a Glia server to connect to
    """

    def __init__(self, parent, host=app.config['LOGIN_SERVER']):
        self.logger = logging.getLogger('e-synapse')
        self.logger.setLevel(app.config['LOG_LEVEL'])

        # Core setup
        self.synapse = parent
        self.souma = Souma.query.get(app.config["SOUMA_ID"])  # The Souma which hosts this Synapse
        self.host = "http://{host}".format(host=host)  # The Glia server to connect to
        self.session = requests.Session()  # Session object to use for requests
        self._peers = dict()
        self._sessions = dict()  # Holds session info for owned Personas (see _get_session(), _set_session()

        # Setup signals
        notification_signals.signal('local-model-changed').connect(self.on_local_model_changed)

        # Test connection to glia-server
        try:
            server_info, errors = self._request_resource("GET", [])
        except requests.ConnectionError, e:
            self.logger.error("Could not establish connection to glia server\n* {}".format(e))
            quit()

        # Register souma if neccessary
        if errors:
            # Check for SOUMA_NOT_FOUND error code in server response
            if server_info and ERROR["SOUMA_NOT_FOUND"](None)[0] in map(itemgetter(0), server_info["meta"]["errors"]):
                if self.souma_register():
                    server_info, errors = self._request_resource("GET", [])
                else:
                    self._log_errors("Error registering Souma", errors)
                    quit()
            else:
                self._log_errors("Error connecting to Glia", errors)
                quit()

        try:
            self.logger.info(
                "\n".join(["{:=^80}".format(" GLIA INFO "),
                    "{:>12}: {} ({})".format("status",
                        server_info["server_status"][0]["status_message"],
                        server_info["server_status"][0]["status_code"]),
                    "{:>12}: {}".format("server id", server_info["server_status"][0]["id"]),
                    "{:>12}: {}".format("personas", server_info["server_status"][0]["personas_registered"]),
                    "{:>12}: {}".format("vesicles", server_info["server_status"][0]["vesicles_stored"])
                ]))
        except KeyError, e:
            self.logger.warning("Received invalid server status: Missing {}".format(e))

    def _get_session(self, persona):
        """
        Return the current session id for persona or create a new session

        Parameters:
            persona (Persona): A persona object to be logged in

        Returns:
            dict A dictionary with keys `id`, containing a session id and
                `timeout` containing a datetime object for the session's
                timeout
        """
        if persona.id in self._sessions:
            return self._sessions[persona.id]
        else:
            return self.persona_login(persona)

    def _set_session(self, persona, session_id, timeout):
        """
        Store a new session id for persona

        Args:
            persona (persona)
            session_id (str): New session id. If set to none, the session is removed
            timeout (str): ISO formatted datetime of session timeout
        """

        if session_id is None:
            del self._sessions[persona.id]
        else:
            to = dateutil_parse(timeout)

            self._sessions[persona.id] = {
                'id': session_id,
                'timeout': to
            }

    def _log_errors(self, msg, errors, level="error"):
        """
        Log a list of errors to the logger

        Args:
            msg(str): A message describing the error source
            errors(list): A list of error messages

        Raises:
            ValueError: If the specified log level is invalid
        """

        if level not in ["debug", "info", "warning", "error"]:
            raise ValueError("Invalid log level {}".format(level))

        call = getattr(self.logger, level)
        call("{msg}:\n{list}".format(msg=msg, list="\n* ".join(str(e) for e in errors)))

    def _keepalive(self, persona):
        """
        Keep @param persona's glia session alive by sending a keep-alive request

        If there is no session yet for persona, she is logged in.
        """

        self.logger.info("Sending keepalive for {}".format(persona))

        session_id = self._get_session(persona)["id"]
        resp, errors = self._request_resource("GET", ["sessions", session_id])

        if errors:
            self._log_errors("Error requesting keepalive for {}".format(persona), errors)

            # Login if session is invalid
            if ERROR["INVALID_SESSION"] in resp['meta']['errors']:
                self.login(persona)
        else:
            session_id = resp['sessions'][0]['id']
            timeout = resp['sessions'][0]['timeout']
            self._set_session(persona, session_id, timeout)

            self._queue_keepalive(persona)

    def _queue_keepalive(self, persona, timeout=900):
        """
        Queue keepalive for persona in @param timeout seconds (default 15 minutes)
        """

        buf = 10  # seconds
        remaining = (self._get_session(persona)['timeout'] - datetime.datetime.utcnow()).seconds - buf
        if (remaining - buf) < 0:
            remaining = 2

        self.logger.debug("Next keepalive for {} queued in {} seconds".format(persona, remaining))

        ping = Greenlet(self._keepalive, persona)
        ping.start_later(remaining)

    def _request_resource(self, method, endpoint, params=None, payload=None):
        """
        Request a resource from the server

        Args:
            method (str): One of "GET", "POST", "PUT", "PATCH", "DELETE"
            endpoint (list): A list of strings forming the path of the API endpoint
            params (dict): Optional parameters to attach to the query strings
            payload (object): Will be attached to the request JSON encoded

        Returns:
            A tuple of two elements:
            [0] (object) The (decoded) response of the server
            [1] (list) A list of error strings specified in the `errors` field of the response
        """
        # Validate params
        HTTP_METHODS_1 = ("GET", "DELETE")
        HTTP_METHODS_2 = ("POST", "PUT", "PATCH")  # have `data` parameter

        if method not in HTTP_METHODS_1 and method not in HTTP_METHODS_2:
            raise ValueError("Invalid request method {}".form(method))

        if payload:
            if not isinstance(payload, dict):
                raise ValueError("Payload must be a dictionary type")
            try:
                payload_json = json.dumps(payload)
            except ValueError, e:
                raise ValueError("Error encoding payload of {}:\n{}".format(self, e))
        else:
            payload_json = None

        # Construct URL
        url_elems = [self.host, "v" + str(API_VERSION)]
        url_elems.extend(endpoint)
        url = "/".join(url_elems) + "/"

        # Make request
        headers = dict()
        errors = list()
        parsing_failed = False

        call = getattr(self.session, method.lower())
        try:
            if method in HTTP_METHODS_1:
                self.logger.debug("{} {}".format(method, url))
                r = call(url, headers=headers, params=params, auth=GliaAuth(souma=self.souma))
            else:
                self.logger.debug("{} {}\n{}".format(method, url, payload_json))
                headers['Content-Type'] = "application/json"
                r = call(url, payload_json,
                    headers=headers, params=params, auth=GliaAuth(souma=self.souma, payload=payload_json))
            r.raise_for_status()
        except requests.exceptions.RequestException, e:
            errors.append(e)

        # Try parsing the response
        resp = None
        try:
            resp = r.json()
            self.logger.debug("Received data:\n{}".format(resp))
        except ValueError, e:
            resp = None
            parsing_failed = True
            errors.append("Parsing JSON failed: {}".format(e))
        except UnboundLocalError:
            parsing_failed = True
            errors.append("No data received")

        error_strings = list()
        if not parsing_failed and "meta" in resp and 'errors' in resp["meta"]:
            for error in resp['meta']['errors']:
                error_strings.append("{}: {}".format(error[0], error[1]))

        # Don't return empty error_strings if parsing server errors has failed, return client-side errors instead
        elif errors:
            error_strings = errors

        # Log all errors
        if errors:
            self.logger.error("{} {} / {} failed.\nParam: {}\nPayload: {}\nErrors:\n* {}".format(
                method, endpoint, url, params, payload_json, "\n* ".join(str(e) for e in errors)))

        return (resp, error_strings)

    def _update_peer_list(self, persona):
        """
        Retrieve current IPs of persona's peers

        Args:
            persona (persona): The Persona whose peers will be located

        Returns:
            list A list of error messages or None
        """

        self.logger.info("Updating peerlist for {}".format(persona))

        contacts = Persona.query.get(persona.id).contacts

        peer_ids = list()  # peers we want to look up
        for p in contacts:
            peer_ids.append(p.id)

        if len(peer_ids) == 0:
            self.logger.info("{} has no peers. Peerlist update cancelled.".format(persona))
        else:
            # ask glia server for peer info
            resp, errors = self._request_resource("GET", ["sessions"], params={'ids': ",".join(peer_ids)})
            # TODO: Remove peers that are no longer online

            if errors:
                self._log_errors("Error updating peer list", errors)
                return errors

            else:
                offline = 0
                for infodict in resp['sessions']:
                    p_id = infodict['id']
                    soumas = infodict['soumas']

                    if soumas:
                        for souma in soumas:
                            # self.souma_discovered.send(self._update_peer_list, message=souma)
                            pass
                    else:
                        offline += 1
                        self.logger.info("No online souma found for {}".format(contacts.get(p_id)))

                self.logger.info("Updated peer list: {}/{} online".format(
                    len(resp["sessions"]) - offline, len(resp["sessions"])))

    def get_persona(self, persona_id):
        """Returns a Persona object for persona_id, loading it from Glia if neccessary

        Args:
            persona_id (String): ID of the required Persona

        Returns:
            Persona: If a record was found
            None: If no record was found
        """
        persona = Persona.query.get(persona_id)

        if persona:
            return persona
        else:
            resp, errors = self.persona_info(persona_id)

            if errors or (resp and "personas" in resp and len(resp["personas"]) == 0):
                self._log_errors("Error requesting Persona {} from server".format(persona_id), errors)
                return None
            else:
                persona = Persona.query.get(persona_id)
                return persona

    def find_persona(self, address):
        """
        Find personas by their email address

        Args:
            address (string): Email address to search for

        Returns:
            list A list of dictionaries containing found profile information

            Keys:
                "persona_id",
                "username",
                "host",
                "port_external",
                "port_internal",
                "crypt_public",
                "sign_public",
                "connectable"
        """

        self.logger.info("Requesting persona record for  '{}'".format(address))

        payload = {
            "email_hash": [sha256(address).hexdigest(), ]
        }

        return self._request_resource("POST", ["personas"], payload=payload)

    def login_all(self):
        """
        Login all personas with a non-empty private key
        """
        persona_set = Persona.query.filter('sign_private != ""').all()
        if len(persona_set) == 0:
            self.logger.warning("No controlled Persona found.")
        else:
            self.logger.info("Logging in {} personas".format(len(persona_set)))
            for p in persona_set:
                self.persona_login(p)

    def myelin_receive(self, recipient_id, interval=None):
        """
        Request Vesicles directed at recipient from Myelin and pass them on to Synapse for handling.

        Parameters:
            recipient_id (String) The ID of the Persona for which to listen
            interval (int) If set to an amount of seconds, the function will repeatedly be called again in this interval
        """
        recipient = Persona.query.get(recipient_id)
        if not recipient:
            self.logger.error("Could not find Persona {}".format(recipient_id))
            return

        self.logger.debug("Updating Myelin of {} at {} second intervals".format(recipient, interval))
        params = dict()

        # Determine offset
        offset = recipient.myelin_offset
        if offset is not None:
            self.logger.debug("Last Vesicle for {} received {} ({})".format(
                recipient, naturaltime(datetime.datetime.utcnow() - offset), offset))
            params["offset"] = str(recipient.myelin_offset)

        resp, errors = self._request_resource("GET", ["myelin", "recipient", recipient.id], params, None)

        if errors:
            self._log_errors("Error receiving from Myelin", errors)
        else:
            for v in resp["vesicles"]:
                vesicle = self.synapse.handle_vesicle(v)
                if vesicle is not None:
                    myelin_modified = iso8601.parse_date(
                        resp["meta"]["myelin_modified"][vesicle.id]).replace(tzinfo=None)
                    if offset is None or myelin_modified > offset:
                        offset = myelin_modified

        # Update recipient's offset if a more recent Vesicle has been received
        if offset is not None:
            if recipient.myelin_offset is None or offset > recipient.myelin_offset:
                recipient.myelin_offset = offset
                session = create_session()
                session.add(recipient)
                session.commit()
                # session.close()

        # Schedule this method to be called in again in interval seconds
        if interval is not None:
            update = Greenlet(self.myelin_receive, recipient_id, interval)
            update.start_later(interval)

    def myelin_store(self, vesicle):
        """
        Store a Vesicle in Myelin

        Parameters:
            vesicle (Vesicle) The vesicle to be stored

        Returns:
            list List of error strings if such occurred
        """
        data = {
            "vesicles": [vesicle.json(), ]
        }

        resp, errors = self._request_resource("PUT", ["myelin", "vesicles", vesicle.id], payload=data)

        if errors:
            self._log_errors("Error transmitting {} to Myelin".format(vesicle), errors)
            return errors
        else:
            self.logger.debug("Transmitted {} to Myelin".format(vesicle))

    def on_local_model_changed(self, sender, message):
        """Check if Personas were changed and call register / unregister method"""
        if message["object_type"] == "Persona":
            persona = Persona.query.get(message["object_id"])

            if message["action"] == "insert":
                self.persona_register(persona)
            elif message["action"] == "update":
                self.logger.warning("Updating Persona profiles in Glia not yet supported")
            elif message["action"] == "delete":
                self.persona_unregister(persona)

    def persona_info(self, persona_id):
        """
        Return a dictionary containing info about a persona, storing it in the db as well.

        Parameters:
            persona_id (str): Persona ID

        Returns:
            tuple 0: persona's public profile and auth_token for authentication
                as a dict, 1: list of errors while requesting from server
        """
        resp, errors = self._request_resource("GET", ["personas", persona_id])

        if errors:
            self._log_errors("Error retrieving Persona profile for ID: {}".format(persona_id), errors)
        else:
            pinfo = resp["personas"][0]

            modified = iso8601.parse_date(pinfo["modified"])

            p = Persona.query.get(persona_id)
            if p is None:
                session = create_session()
                try:
                    p = Persona(
                        _stub=True,
                        id=persona_id,
                        username=pinfo["username"],
                        email=pinfo.get("email"),
                        modified=modified,
                        sign_public=pinfo["sign_public"],
                        crypt_public=pinfo["crypt_public"]
                    )
                    session.add(p)
                    session.commit()
                    self.logger.info("Loaded {} from Glia server".format(p))
                except KeyError, e:
                    self.logger.warning("Missing key in server response for storing new Persona: {}".format(e))
                    errors.append("Missing key in server response for storing new Persona: {}".format(e))
                except:
                    session.rollback()
                    raise
                finally:
                    # session.close()
                    pass
            else:
                # TODO: Update persona info
                pass

        return resp, errors

    def persona_login(self, persona):
        """
        Login a persona on the server, register if not existing. Start myelinated reception if activated.

        Returns:
            str -- new session id
        """
        # Check current state
        if persona.id in self._sessions:
            return self._get_session(persona)["session_id"]

        # Obtain auth token
        info, errors = self._request_resource("GET", ["personas", persona.id])
        if errors:
            self._log_errors("Error logging in", errors)

            # Register persona if not existing
            if ERROR["OBJECT_NOT_FOUND"](None)[0] in map(itemgetter(0), info["meta"]["errors"]):
                errors = self.persona_register(persona)

                if errors:
                    self.logger.error("Failed logging in / registering {}.".format(persona))
                    return None
                else:
                    session = self._get_session(persona)
                    return session["id"]
        try:
            auth = info["personas"][0]["auth"]
        except KeyError, e:
            self.logger.warning("Server sent invalid response: Missing `{}` field.".format(e))
            return None

        # Send login request
        data = {
            "personas": [{
                "id": persona.id,
                'auth_signed': persona.sign(auth),
                'reply_to': app.config["SYNAPSE_PORT"]
            }]
        }
        resp, errors = self._request_resource("POST", ["sessions"], payload=data)

        # Read response
        if errors:
            self._log_errors("Login failed", errors)
            return None
        else:
            session_id = resp["sessions"][0]['id']
            timeout = resp["sessions"][0]['timeout']

            self.logger.info("Persona {} logged in until {}".format(persona, timeout))
            self._set_session(persona, session_id, timeout)
            self._queue_keepalive(persona)
            self._update_peer_list(persona)
            if app.config["ENABLE_MYELIN"]:
                self.myelin_receive(persona.id, interval=app.config["MYELIN_POLLING_INTERVAL"])

            return {
                "id": session_id,
                "timeout": timeout
            }

    def persona_logout(self, persona):
        """
        Terminate persona's session on the host

        Parameters:
            persona -- persona to be logged out

        Returns:
            dict -- error_name:error_message
        """
        self.logger.info("Logging out {}".format(persona))
        ses = self._get_session(persona)
        resp, errors = self._request_resource("DELETE", ["personas", persona.id, ses["id"]])

        if errors:
            self._log_errors("Error logging out", errors)
            return errors
        else:
            self._set_session(persona, None, None)
            self.logger.info("Logged out {}".format(persona))

    def persona_register(self, persona):
        """
        Register a persona on the server

        Parameters:
            persona -- persona to be registered

        Returns:
            list -- error messages
        """
        # Create request
        data = {
            "personas": [{
                'persona_id': persona.id,
                'username': persona.username,
                'modified': persona.modified.isoformat(),
                'email_hash': persona.get_email_hash(),
                'sign_public': persona.sign_public,
                'crypt_public': persona.crypt_public,
                'reply_to': app.config['SYNAPSE_PORT']
            }, ]
        }

        response, errors = self._request_resource("PUT", ["personas", persona.id], payload=data)

        if errors:
            self._log_errors("Error creating glia profile for {}".format(persona), errors)
            return errors

        # Evaluate response
        try:
            session_id = response['sessions'][0]['id']
            timeout = response['sessions'][0]['timeout']
        except KeyError, e:
            return ["Invalid server response: Missing key `{}`".format(e)]

        self.logger.info("Registered {} with server.".format(persona))
        self._set_session(persona, session_id, timeout)
        self._update_peer_list(persona)
        self._queue_keepalive(persona)
        if app.config["ENABLE_MYELIN"]:
            self.myelin_receive(persona.id, interval=app.config["MYELIN_POLLING_INTERVAL"])

    def persona_unregister(self, persona):
        """
        Remove persona's data from the glia server

        Parameters:
            persona (persona):persona to be unregistered

        Returns:
            dict error_name:error_message
        """

        self.logger.info("Unregistering {}".format(persona))

        response, errors = self._request_resource("DELETE", ["personas", persona.id])

        if errors:
            self._log_errors("Error unregistering persona {}".format(persona), errors)
        else:
            self.logger.info("Unregisterd persona {} from Glia server:\n{}".format(persona, response))

    def souma_register(self):
        """
        Register this Souma with the glia server

        Returns:
            bool -- True if successful
        """

        self.logger.info("Registering local {} with Glia-server".format(self.souma))

        data = {
            "soumas": [{
                "id": self.souma.id,
                "crypt_public": self.souma.crypt_public,
                "sign_public": self.souma.sign_public,
            }, ]
        }

        response, errors = self._request_resource("POST", ["soumas"], payload=data)

        if errors:
            self._log_errors("Registering {} failed".format(self.souma), errors)
            return False
        else:
            self.logger.info("Successfully registered {} with server".format(self.souma))
            return True

    def shutdown(self):
        """
        Terminate connections and logout
        """
        for p in self._sessions:
            self.persona_logout(p['id'])



import logging
import blinker

from web_ui import db, app

from sqlalchemy.orm import sessionmaker

ERROR = {
    "MISSING_MESSAGE_TYPE": (1, "No message type found."),
    "MISSING_PAYLOAD": (2, "No data payload found."),
    "OBJECT_NOT_FOUND": lambda name: (3, "Object does not exist: ".format(name)),
    "MISSING_KEY": lambda name: (4, "Missing data for this request: {}".format(name)),
    "INVALID_SIGNATURE": (5, "Invalid signature."),
    "INVALID_SESSION": (6, "Session invalid. Please re-authenticate."),
    "DUPLICATE_ID": lambda id: (7, "Duplicate ID: {}".format(id)),
    "SOUMA_NOT_FOUND": lambda id: (8, "Souma not found: {}".format(id)),
    "MISSING_PARAMETER": lambda name: (9, "Missing HTTP parameter: {}".format(name)),
}

# Setup Blinker namespace
notification_signals = blinker.Namespace()

# Setup logger namespace
logger = logging.getLogger('nucleus')

# Source formatting helper
source_format = lambda address: None if address is None else \
    "{host}:{port}".format(host=address[0], port=address[1])

# Possible states of stars
STAR_STATES = {
    -2: (-2, "deleted"),
    -1: (-1, "unavailable"),
    0: (0, "published"),
    1: (1, "draft"),
    2: (2, "private"),
    3: (3, "updating")
}

# Possible states of planets
PLANET_STATES = {
    -1: (-1, "unavailable"),
    0: (0, "published"),
    1: (1, "private"),
    2: (2, "updating")
}

# Possible states of 1ups
ONEUP_STATES = {
    -1: "disabled",
    0: "active",
    1: "unknown author"
}

CHANGE_TYPES = ("insert", "update", "delete")


class InvalidSignatureError(Exception):
    """Throw this error when a signature fails authenticity checks"""
    pass


class PersonaNotFoundError(Exception):
    """Throw this error when the Persona profile specified for an action is not available"""
    pass


class UnauthorizedError(Exception):
    """Throw this error when the active Persona is not authorized for an action"""
    pass


class VesicleStateError(Exception):
    """Throw this error when a Vesicle's state does not allow for an action"""
    pass


# Import at bottom to avoid circular imports
# Import all models to allow querying db binds
from nucleus.models import *
from vesicle import Vesicle

# _Session is a custom sessionmaker that returns a session prefconfigured with the
# model bindings from Nucleus
_Session = sessionmaker(bind=db.get_engine(app))


def create_session():
    """Return a session to be used for database connections

    Returns:
        Session: SQLAlchemy session object
    """
    # Produces integrity errors!
    # return _Session()

    # db.session is managed by Flask-SQLAlchemy and bound to a request
    return db.session



import lxml.html
from gensim import utils
from gevent import spawn_later
import urllib
from readability import Document


def _clean_attrib(node):
    for n in node:
        _clean_attrib(n)
    node.attrib.clear()


def remove_html(text):
    tree = lxml.html.fromstring(text)
    cleaner = lxml.html.clean.Cleaner(style=True)
    cleaner.clean_html(tree)
    _clean_attrib(tree)

    return lxml.html.tostring(tree, encoding='unicode', pretty_print=True,
                              method='text')


def tokenize(text):
    text = remove_html(text)

    return [token.encode('utf8')
            for token in utils.tokenize(text, lower=True, errors='ignore')
            if 2 <= len(token) <= 15 and not token.startswith('_')]


def get_site_content(link):
    # request link content
    req = urllib.urlopen(link)
    page = req.read()

    # extract the  (most likely) main content
    doc = Document(page, url=link)
    content = doc.summary(html_partial=True)

    return remove_html(content)


def repeated_func_schedule(time, func):
    spawn_later(0, func)
    spawn_later(time, repeated_func_schedule, time, func)



import logging
import json

from dateutil.parser import parse as dateutil_parse
from uuid import uuid4

from nucleus import create_session, notification_signals, PersonaNotFoundError, UnauthorizedError, VesicleStateError, CHANGE_TYPES
from nucleus.models import Persona, Star, Planet, Starmap, Group
from nucleus.vesicle import Vesicle
from synapse.electrical import ElectricalSynapse
from web_ui import app

# These are Vesicle options which are recognized by this Synapse
ALLOWED_MESSAGE_TYPES = [
    "object",
    "object_request",
]

OBJECT_TYPES = ("Star", "Planet", "Persona", "Starmap", "Group")


class Synapse():
    """
    A Synapse reacts to local changes in the database and transmits
    them to each Persona's peers using the Myelin API. It also keeps
    Glia up to date on all Persona's managed by this Souma.
    """

    def __init__(self):
        self.logger = logging.getLogger('synapse')
        self.logger.setLevel(app.config['LOG_LEVEL'])

        # Core setup
        self.starmap = Starmap.query.get(app.config['SOUMA_ID'])

        # Connect to glia
        self.electrical = ElectricalSynapse(self)
        self.electrical.login_all()

        # Connect to nucleus
        self._connect_signals()

    def _connect_signals(self):
        """
        Connect to Blinker signals which are registered in nucleus.__init__
        """

        signal = notification_signals.signal

        signal('local-model-changed').connect(self.on_local_model_change)
        signal('new-contact').connect(self.on_new_contact)
        signal('request-objects').connect(self.on_request_objects)
        signal('group-created').connect(self.on_group_created)

    def _distribute_vesicle(self, vesicle, recipients=None):
        """
        Encrypt, sign and distribute vesicle to `recipients` using Myelin

        Args:
            vesicle (Vesicle): The message to transmit.
            recipients (List): List of Persona objects. If recipients is not
                empty, the Vesicle is encrypted and the key is transmitted for
                this list of Personas.

        Returns:
            Vesicle: The (signed and encrypted) Vesicle object
        """
        self.logger.debug("Distributing {} to {} recipients {}".format(
            vesicle,
            len(recipients) if recipients is not None else "0",
            "via Myelin" if app.config["ENABLE_MYELIN"] else ""))

        if not hasattr(vesicle, "author"):
            raise ValueError("Can't send Vesicle without defined author")

        if vesicle.encrypted():
            keycrypt = json.loads(vesicle.keycrypt)

            # First remove everyone from keycrypt that is not a current recipient
            keycrypt = json.loads(vesicle.keycrypt)
            remove_recipients = set(keycrypt.keys()) - set([r.id for r in recipients])
            for recipient_id in remove_recipients:
                if recipient_id != vesicle.author_id:  # Don't remove author!
                    del keycrypt[recipient_id]
            vesicle.keycrypt = json.dumps(keycrypt)

            # Then add the new recipients
            vesicle.add_recipients(recipients)

            self.logger.debug("{} was already encrypted: Modified keycrypt.".format(vesicle))
        else:
            vesicle.encrypt(recipients=recipients)

        try:
            vesicle.sign()
        except VesicleStateError:
            self.logger.info("{} was already signed".format(vesicle))

        if app.config["ENABLE_MYELIN"]:
            self.electrical.myelin_store(vesicle)

        return vesicle

    def _find_source(self, obj):
        """Return a list of possible sources for object.

        A Persona qualifies as source if they have obj in their starmaps,
        have a controlled Persona as a contact.

        Args:
            obj (Star, Planet or Starmap): Object to find a source for

        Returns:
            list: Possible sources
        """
        # Return True if at least one of controlled personas is a contact of p
        connected_to = lambda p: len(p.contacts.filter(Persona.crypt_private != "")) > 0

        sources = list()
        if isinstance(obj, Star):
            if connected_to(obj.author):
                sources.append(obj.author)

            for starmap in obj.starmaps:
                if connected_to(starmap.author):
                    sources.append(starmap.author)

        elif isinstance(obj, Planet):
            for star in obj.stars:
                if connected_to(star.author):
                    sources.append(star.author)

                for starmap in star.starmaps:
                    if connected_to(starmap.author):
                        sources.append(starmap.author)

        elif isinstance(obj, Starmap):
            if connected_to(obj.author):
                sources.append(obj.author)

        # TODO: Sort by last seen
        return sources

    def _log_errors(self, msg, errors, level="error"):
        """
        Log a list of errors to the logger

        Args:
            msg(str): A message describing the error source
            errors(list): A list of error messages

        Raises:
            ValueError: If the specified log level is invalid
        """

        if level not in ["debug", "info", "warning", "error"]:
            raise ValueError("Invalid log level {}".format(level))

        call = getattr(self.logger, level)
        call("{msg}:\n{list}".format(msg=msg, list="\n* ".join(str(e) for e in errors)))

    def handle_object(self, vesicle, reader_persona, session):
        """
        Handle received object updates by verifying the request and calling
        an appropriate handler

        Args:
            vesicle (Vesicle): Vesicle containing the object changeset
            reader_persona (Persona): Persona used for decrypting the Vesicle
        """
        # Validate response
        errors = list()

        try:
            action = vesicle.data["action"]
            object_type = vesicle.data["object_type"]
            obj = vesicle.data["object"]
            author = vesicle.author
        except KeyError, e:
            errors.append("Missing key: {}".format(e))

        if object_type not in OBJECT_TYPES:
            errors.append("Unknown object type: {}".format(object_type))

        if action not in CHANGE_TYPES:
            errors.append("Unknown action type '{}'".format(action))

        if errors:
            self.logger.error("Malformed object received\n{}".format("\n".join(errors)))
        else:
            handler = getattr(self, "object_{}".format(action))
            new_obj = handler(author, reader_persona, object_type, obj, session)

            if new_obj is not None:
                vesicle.handled = True
                new_obj.vesicles.append(vesicle)

                session.add(new_obj)
                session.add(vesicle)
                session.commit()

    def handle_object_request(self, vesicle, reader_persona, session):
        """
        Act on received object requests by sending the object in question back

        Args:
            vesicle (Vesicle): Vesicle containing metadata about the object
        """

        # Validate vesicle
        errors = []
        object_id = None
        object_type = None

        try:
            object_id = vesicle.data["object_id"]
            object_type = vesicle.data["object_type"]
            recipient = vesicle.author
        except KeyError, e:
            errors.append("missing ({})".format(vesicle, e))

        if object_type not in OBJECT_TYPES:
            errors.append("invalid object_type: {}".format(object_type))

        if errors:
            self._log_errors("Received invalid object request", errors)
        else:
            # Load object
            obj_class = globals()[object_type]
            obj = obj_class.query.get(object_id)

            if obj is None:
                self.logger.error("Requested object <{type} {id}> not found".format(
                    type=object_type, id=object_id[:6]))

            elif hasattr(obj, "author") and recipient not in obj.author.contacts:
                    self.logger.info("Requested {} not published for request author {}".format(obj, recipient))
            elif isinstance(obj, Persona) and recipient not in obj.contacts:
                    self.logger.info("Requested {} does not have requesting {} as a contact.".format(obj, recipient))
            else:
                for v in obj.vesicles:
                    # Send response
                    # Modified vesicles (re-encrypted) don't get saved to DB
                    self._distribute_vesicle(v, recipients=[recipient, ])
                self.logger.info("Sent {}'s {} vesicles to {}".format(
                    obj, len(obj.vesicles), recipient
                ))

        vesicle.handled = True
        session.add(vesicle)
        session.commit()

    def handle_vesicle(self, data):
        """
        Parse received vesicles and call handler

        Args:
            data (String): JSON encoded Vesicle

        Returns:
            Vesicle: The Vesicle that was decrypted and loaded
            None: If no Vesicle could be loaded
        """

        try:
            vesicle = Vesicle.read(data)
        except PersonaNotFoundError, e:
            self.logger.info("Received Vesicle from unknown Persona, trying to retrieve Persona info.")
            resp, errors = self.electrical.persona_info(e[0])
            if errors:
                self.logger.warning("Could not retrieve unknown Persona from server:\n{}".format(", ".join(errors)))
                return
            else:
                vesicle = Vesicle.read(data)

        if vesicle is None:
            self.logger.error("Failed handling Vesicle due to decoding error")
        else:
            old_vesicle = Vesicle.query.get(vesicle.id)
            if old_vesicle is not None:
                vesicle = old_vesicle

            # Vesicle is loaded and has not yet been handled, start processing..
            session = create_session()

            # Decrypt if neccessary
            import keyczar
            if not vesicle.decrypted():
                try:
                    reader_persona = vesicle.decrypt()
                except UnauthorizedError:
                    self.logger.info("Not authorized to decrypt {}".format(vesicle))
                except keyczar.errors.InvalidSignatureError:
                    self.logger.warning("Failed decrypting {}: id={} h={}".format(vesicle, vesicle.id, vesicle._get_hashcode()))
                    return vesicle

            if old_vesicle is None:
                session.add(vesicle)
                session.commit()

            if not vesicle.decrypted():
                self.logger.debug("{} has encrypted payload.".format(vesicle))
            else:
                self.logger.info("{} has payload:\n{}".format(vesicle, json.dumps(vesicle.data, indent=2)))

            # Call handler depending on message type
            try:
                if vesicle is not None and not vesicle.handled and vesicle.message_type in ALLOWED_MESSAGE_TYPES:
                    handler = getattr(self, "handle_{}".format(vesicle.message_type))
                    try:
                        handler(vesicle, reader_persona, session)
                    except UnauthorizedError, e:
                        self.logger.error("Error handling {}: {}".format(vesicle, e))
            except:
                session.rollback()
                raise
            finally:
                session.flush()

        return vesicle

    def object_insert(self, author, recipient, object_type, obj, session):
        # Handle answer
        obj_class = globals()[object_type]
        missing_keys = obj_class.validate_changeset(obj)
        if len(missing_keys) > 0:
            raise KeyError("Missing '{}' for creating {} from changeset".format(
                ", ".join(missing_keys), obj_class.__name__))

        if hasattr(obj, "author_id") and author.id != obj["author_id"]:
            raise UnauthorizedError(
                "Received object_insert Vesicle author {} does not match object author [{}]".format(
                    author, obj["author_id"][:6]))

        o = obj_class.query.get(obj["id"])
        if o is None or (hasattr(o, "get_state") and o.get_state() == -1) or (isinstance(o, Persona) and o._stub is True):
            o = obj_class.create_from_changeset(obj, stub=o, update_sender=author, update_recipient=recipient)
            session.add(o)
            if isinstance(o, Persona):
                o.stub = False
            else:
                o.set_state(0)
            self.logger.info("Inserted new {}".format(o))

        return o

    def object_update(self, author, recipient, object_type, obj, session):
        # Verify message
        obj_class = globals()[object_type]
        missing_keys = obj_class.validate_changeset(obj, update=True)
        if len(missing_keys) > 0:
            raise KeyError("Missing '{}' for updating {} from changeset".format(obj_class.__name__))

        obj_modified = dateutil_parse(obj["modified"])

        # Retrieve local object copy
        o = obj_class.query.get(obj["id"])

        if o is None:
            self.logger.warning("Received update for unknown <{} [{}]>".format(object_type, obj["id"][:6]))

            o = obj_class(id=obj["id"])
            o.set_state(-1)
            session.add(o)

            self.request_object(
                object_type=object_type,
                object_id=obj["id"],
                author=recipient,
                recipient=author,
                session=session)
        else:
            if o.authorize("update", author.id):
                if o.modified <= obj_modified or (hasattr(o, "_stub") and o._stub is True):
                    o.update_from_changeset(obj, update_sender=author, update_recipient=recipient)
                    if isinstance(o, Persona):
                        o.stub = False
                    else:
                        o.set_state(0)
                    session.add(o)
                    self.logger.info("Applied update for {}".format(o))
                else:
                    self.logger.info("Received obsolete update ({})".format(obj))
            else:
                self.logger.warning("{} is not authorized to update {} - update canceled.".format(author, o))

        return o

    def object_delete(self, author, recipient, object_type, obj, session):
        # Verify message
        obj_class = globals()[object_type]
        for k in ["id", "modified"]:
            if k not in obj.keys():
                raise KeyError("Missing '{}' for deleting {} from changeset".format(k, obj_class.__name__))

        # Get the object's class from globals
        o = obj_class.query.get(obj["id"])

        if o is None:
            self.logger.info("Request to delete unknown <{} [{}]>".format(object_type, obj["id"]))
        else:
            if o.authorize("delete", author.id):
                if hasattr(o, "set_state"):
                    o.set_state(-2)
                    o.text = None
                    session.add(o)
                    self.logger.info("{} marked deleted".format(o))
                else:
                    name = str(o)
                    session.delete(o)
                    o = None
                    self.logger.info("Permanently deleted {}".format(name))
            else:
                self.logger.warning("Object deletion not authorized!")
        return o

    def on_new_contact(self, sender, message):
        for k in["new_contact_id", "author_id"]:
            if k not in message:
                raise KeyError("Missing message parameter '{}'".format(k))

        author = Persona.query.get(message["author_id"])
        recipient = Persona.query.get(message["new_contact_id"])

        session = create_session()

        if recipient._stub is True:
            self.logger.info("Requesting profile of new contact {}".format(recipient))
            self.request_object("Persona", message["new_contact_id"], author, recipient, session)

        session.commit()

    def on_local_model_change(self, sender, message):
        """
        React to model changes reported from the web-ui by transmitting
        appropriate messages to peers

        Args:
            sender(object): Sender of the Blinker signal
            message(dict): Changeset containing keys in CHANGESET_REQUIRED_FIELDS

        Raises:
            KeyError: Missing key from Changeset
            ValueError: Changeset contains illegal value
        """
        CHANGESET_REQUIRED_FIELDS = ["author_id", "action", "object_id", "object_type"]

        # Verify changeset
        for k in CHANGESET_REQUIRED_FIELDS:
            if k not in message.keys():
                raise KeyError("Missing key: '{}'".format(k))

        if message["action"] not in CHANGE_TYPES:
            raise ValueError("Unknown action: '{}'".format(message["action"]))

        if message["object_type"] not in OBJECT_TYPES:
            raise ValueError("Object type {} not supported".format(message["object_type"]))

        # Get the object's class from globals
        obj_class = globals()[message["object_type"]]
        obj = obj_class.query.get(message["object_id"])

        if obj is None:
            # TODO: Use different exception type
            raise Exception("Could not find {} {}".format(obj_class, message["object_id"]))

        author = Persona.query.get(message["author_id"])

        # Send Vesicle
        data = dict({
            "action": message["action"],
            "object_type": message["object_type"]
        })

        if message["action"] == "insert":
            data["object"] = obj.export()
        elif message["action"] == "update":
            data["object"] = obj.export(update=True)
        elif message["action"] == "delete":
            data["object"] = {
                "id": obj.id,
                "modified": obj.modified.isoformat()
            }

        vesicle = Vesicle(
            id=uuid4().hex,
            message_type="object",
            data=data,
            author=author,
            handled=True
        )
        obj.vesicles.append(vesicle)

        session = create_session()

        try:
            session.add(vesicle)
            session.commit()
        except:
            session.rollback()
            raise
        else:
            self.logger.info("Local {} changed: Distributing {}".format(obj, vesicle))
            vesicle = self._distribute_vesicle(vesicle, recipients=author.contacts.all())

        session.add(vesicle)
        session.commit()

    def on_request_objects(self, sender, message):
        """React to request-objects signal by queuing a request

        Args:
            sender (object): Sender of the signal
            message (dict): Contains keys
                type -- object type as defined in OBJECT_TYPES
                id -- 32 byte object ID
                author_id -- (optional) author of the request
                recipient_id -- (optional) recipient of the request
        """
        try:
            object_type = message["type"]
            object_id = message["id"]
        except KeyError, e:
            self.logger.warning("Missing request parameter '{}'".format(e))
            return

        session = create_session()

        if "author_id" in message:
            author = self.electrical.get_persona(message["author_id"])
        else:
            author = None

        if "recipient_id" in message:
            recipient = self.electrical.get_persona(message["recipient_id"])
        else:
            recipient = None

        try:
            self.request_object(object_type, object_id, author, recipient, session)
        except:
            session.rollback()
            raise
        finally:
            session.flush()

    def on_group_created(self, sender, message):
        # TODO: Implement how to react on blinker notification
        pass

    def request_object(self, object_type, object_id, author, recipient, session):
        """
        Send a request for an object to a Persona

        Args:
            object_type (String): capitalized class name of the object
            object_id (String): 32 byte object ID
            author (Persona): Author of this request
            recipient (Persona): Persona to request this object from

        Raises:
            UnauthorizedError: If no source can be found that has one of the controlled Personas as a contact
        """
        obj_class = globals()[object_type]
        obj = obj_class.query.get(object_id)

        # Find a source if none is specified
        if recipient is None:
            sources = self._find_source(obj)
            recipient = sources[0] if len(sources) > 0 else None

        if recipient is None:
            self.logger.error("No known source for <{} [{}]>".format(object_type, object_id[:6]))
        else:
            # Try and find a contact of source we can use as this request's author
            if author is None:
                author = recipient.contacts.filter_by(Persona.crypt_private!="").first()

            # Abort if no author was found
            if author is None:
                raise UnauthorizedError("Could not find a source for {} who you are contacts with")

            self.logger.info("Requesting <{object_type} {object_id}> as {author} from {source}".format(
                object_type=object_type, object_id=object_id[:6], author=author, source=recipient))

            data = {
                "object_type": object_type,
                "object_id": object_id
            }

            vesicle = Vesicle(
                id=uuid4().hex,
                message_type="object_request",
                author=author,
                data=data
            )
            session.add(vesicle)
            session.commit()
            session.refresh(vesicle)

            vesicle = self._distribute_vesicle(vesicle, recipients=[recipient])
            session.add(vesicle)
            session.commit()

    def shutdown(self):
        self.pool.kill()



"""
Script to install Souma on OsX, Windows, and Unix

Usage:
    python setup.py py2app
"""
import ez_setup
ez_setup.use_setuptools()

import sys
import shutil
from setuptools import setup

APP = ['run.py']


""" Platform specific options """
if sys.platform == 'darwin':
    extra_options = dict(
        app=['run.py'],
    )
elif sys.platform == 'win32':
    extra_options = dict(
        setup_requires=['py2exe'],
        app=APP,
    )
else:
    extra_options = dict(
        scripts=APP)

print 'Deleting build dirs...'
try:
    shutil.rmtree('../build')
    shutil.rmtree('../dist')
except OSError:
    pass

setup(
    name="Souma",
    version="0.2.1",
    author="Cognitive Networks Group",
    author_email="cognitive-networks@googlegroups.com",
    url="https://github.com/ciex/souma/",
    scripts=["run.py", "set_hosts.py"],
    packages=["nucleus", "web_ui", "synapse", "astrolab"],
    license="Apache License 2.0",
    description="A Cognitive Network for Groups",
    long_description=open("README").read(),
    install_requires=open("requirements.txt").read(),
    **extra_options
)



from gensim.models import LdaModel
from gensim.corpora import Dictionary
from astrolab.helpers import tokenize


class TopicModel(object):

    num_topics = 200

    def __init__(self, lda_file, dic_file):
        self.lda_model = LdaModel.load(lda_file)
        self.dictionary = Dictionary.load_from_text(dic_file)

    def get_topics_text(self, text):
        clear_content = tokenize(text)
        bow = self.dictionary.doc2bow(clear_content)

        return self.get_topics_bow(bow)

    def get_topics_bow(self, bow):
        topics = self.lda_model[bow]
        topic_vec = [0] * self.num_topics
        for topic in topics:
            topic_vec[topic[0]] = topic[1]

        return topic_vec



import datetime
import json
import iso8601

from base64 import b64encode, b64decode
from flask import url_for, session
from hashlib import sha256
from keyczar.keys import RsaPrivateKey, RsaPublicKey
from sqlalchemy import ForeignKey
from uuid import uuid4

from nucleus import ONEUP_STATES, STAR_STATES, PLANET_STATES, \
    PersonaNotFoundError, UnauthorizedError, notification_signals, CHANGE_TYPES
from web_ui import app, db
from web_ui.helpers import epoch_seconds

request_objects = notification_signals.signal('request-objects')


class Serializable():
    """ Make SQLAlchemy models json serializable

    Attributes:
        _insert_required: Default attributes to include in export
        _update_required: Default attributes to include in export with update=True
    """
    id = None
    modified = None

    _insert_required = ["id", "modified"]
    _update_required = ["id", "modified"]

    def authorize(self, action, author_id=None):
        """Return True if this object authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        if action not in CHANGE_TYPES:
            return False
        return True

    def export(self, exclude=[], update=False):
        """Return this object as a dict.

        Args:
            update (Bool): Export only attributes defined in `self._update_required`

        Returns:
            Dict: The serialized object

        Raises:
            KeyError: If a key was not found
        """
        attr_names = self._update_required if update is True else self._insert_required
        attr_names = [a for a in attr_names if a not in exclude]

        return {attr: str(getattr(self, attr)) for attr in attr_names}

    def json(self, update=False):
        """Return this object JSON encoded.

        Args:
            update (Boolean): (optiona) See export docstring

        Returns:
            Str: JSON-encoded serialized instance
        """
        return json.dumps(self.export(update=update), indent=4)

    @classmethod
    def validate_changeset(cls, changeset, update=False):
        """Check whether changeset contains all keys defined as required for this class.

        Args:
            changeset(dict): See created_from_changeset, update_from_changeset
            update(Bool): If True use cls._update_required instead of cls._insert_required

        Returns:
            List: Missing keys
        """
        required_keys = cls._update_required if update else cls._insert_required
        missing = list()

        for k in required_keys:
            if k not in changeset.keys():
                missing.append(k)
        return missing

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new instance from a changeset.

        Args:
            changeset (dict): Dictionary of model values. Requires all keys
                defined in cls._insert_required with class-specific values.
            stub (Serializable): (Optional) model instance whose values will be
                overwritten with those defined in changeset.
            update_sender (Persona): (Optional) author of this changeset. Will be
                used as recipient of subsequent object requests.
            update_recipient (Persona): (Optional) recipient of this changeset.
                Will be used as sender of subsequent object requests.

        Returns:
            Serializable: Instance created from changeset

        Raises:
            KeyError: Missing key in changeset
            TypeError: Argument has wrong type
            ValueError: Argument value cannot be processed
        """
        raise NotImplementedError()

    def update_from_changeset(self, changeset, update_sender=None, update_recipient=None):
        """Update self with new values in changeset

        Args:
            changeset (dict): Dictionary of model values. Requires all keys
                defined in self._update_required with class-specific values.
            update_sender (Persona): (Optional) author of this changeset. Will be
                used as recipient of subsequent object requests.
            update_recipient (Persona): (Optional) recipient of this changeset.
                Will be used as sender of subsequent object requests.

        Returns:
            Serializable: Updated instance

        Raises:
            KeyError: Missing key in changeset
            TypeError: Argument has wrong type
            ValueError: Argument value cannot be processed
        """
        raise NotImplementedError()

#
# Setup follower relationship on Persona objects
#

t_contacts = db.Table(
    'contacts',
    db.Column('left_id', db.String(32), db.ForeignKey('persona.id')),
    db.Column('right_id', db.String(32), db.ForeignKey('persona.id'))
)

t_persona_vesicles = db.Table(
    'persona_vesicles',
    db.Column('persona_id', db.String(32), db.ForeignKey('persona.id')),
    db.Column('vesicle_id', db.String(32), db.ForeignKey('vesicle.id'))
)


class Persona(Serializable, db.Model):
    """A Persona represents a user profile

    Attributes:
        _insert_required: Attributes that are serialized
        id: 32 byte ID generated by uuid4().hex
        username: Public username of the Persona, max 80 bytes
        email: An email address, max 120 bytes
        crypt_private: Private encryption RSA key, JSON encoded KeyCzar export
        crypt_public: Public encryption RSA key, JSON encoded KeyCzar export
        sign_private: Private signing RSA key, JSON encoded KeyCzar export
        sign_public: Public signing RSA key, JSON encoded KeyCzar export
        modified: Last time this Persona object was modified, defaults to now
        contacts: List of this Persona's contacts
        vesicles: List of Vesicles that describe this Persona object
        profile: Starmap containing this Persona's profile page
        index: Starmap containing all Star's this Persona publishes to its contacts
        myelin_offset: Datetime of last request for Vesicles sent to this Persona

    """

    __tablename__ = "persona"

    _insert_required = ["id", "username", "email", "crypt_public",
        "sign_public", "modified", "profile_id", "index_id", "contacts"]
    _update_required = ["id", "username", "email", "profile_id", "index_id", "modified", "contacts"]

    _stub = db.Column(db.Boolean, default=False)
    id = db.Column(db.String(32), primary_key=True)
    username = db.Column(db.String(80))
    email = db.Column(db.String(120))
    crypt_private = db.Column(db.Text)
    crypt_public = db.Column(db.Text)
    sign_private = db.Column(db.Text)
    sign_public = db.Column(db.Text)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())

    contacts = db.relationship(
        'Persona',
        secondary='contacts',
        lazy="dynamic",
        primaryjoin='contacts.c.left_id==persona.c.id',
        secondaryjoin='contacts.c.right_id==persona.c.id')

    vesicles = db.relationship(
        'Vesicle',
        secondary='persona_vesicles',
        primaryjoin='persona_vesicles.c.persona_id==persona.c.id',
        secondaryjoin='persona_vesicles.c.vesicle_id==vesicle.c.id')

    profile_id = db.Column(db.String(32), db.ForeignKey('starmap.id'))
    profile = db.relationship('Starmap', primaryjoin='starmap.c.id==persona.c.profile_id')

    index_id = db.Column(db.String(32), db.ForeignKey('starmap.id'))
    index = db.relationship('Starmap', primaryjoin='starmap.c.id==persona.c.index_id')

    # Group membership

    # groups = db.relationship(
    # 'Group',
    # secondary='groups',
    # primaryjoin='groups.c.left_id==persona.c.id',
    # secondaryjoin='groups.c.right_id==group.c.id')

    # Myelin offset stores the date at which the last Vesicle receieved from Myelin was created
    myelin_offset = db.Column(db.DateTime)

    def __repr__(self):
        return "<{} [{}]>".format(str(self.username), self.id[:6])

    def authorize(self, action, author_id=None):
        """Return True if this Persona authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        if Serializable.authorize(self, action, author_id=author_id):
            return (self.id == author_id)
        return False

    def controlled(self):
        """
        Return True if this Persona has private keys attached
        """
        if self.crypt_private is not None and self.sign_private is not None:
            return True
        else:
            return False

    @staticmethod
    def list_controlled():
        return Persona.query.filter('Persona.sign_private != ""')

    def get_email_hash(self):
        """Return sha256 hash of this user's email address"""
        return sha256(self.email).hexdigest()

    def get_absolute_url(self):
        return url_for('persona', id=self.id)

    def export(self, update=False):
        data = Serializable.export(self, exclude=["contacts", ], update=update)

        data["contacts"] = list()
        for contact in self.contacts:
            data["contacts"].append({
                "id": contact.id,
            })

        return data

    def generate_keys(self, password):
        """ Generate new RSA keypairs for signing and encrypting. Commit to DB afterwards! """

        # TODO: Store keys encrypted
        rsa1 = RsaPrivateKey.Generate()
        self.sign_private = str(rsa1)
        self.sign_public = str(rsa1.public_key)

        rsa2 = RsaPrivateKey.Generate()
        self.crypt_private = str(rsa2)
        self.crypt_public = str(rsa2.public_key)

    def encrypt(self, data):
        """ Encrypt data using RSA """

        key_public = RsaPublicKey.Read(self.crypt_public)
        return b64encode(key_public.Encrypt(data))

    def decrypt(self, cypher):
        """ Decrypt cyphertext using RSA """

        cypher = b64decode(cypher)
        key_private = RsaPrivateKey.Read(self.crypt_private)
        return key_private.Decrypt(cypher)

    def sign(self, data):
        """ Sign data using RSA """

        key_private = RsaPrivateKey.Read(self.sign_private)
        signature = key_private.Sign(data)
        return b64encode(signature)

    def verify(self, data, signature_b64):
        """ Verify a signature using RSA """

        signature = b64decode(signature_b64)
        key_public = RsaPublicKey.Read(self.sign_public)
        return key_public.Verify(data, signature)

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """See Serializable.create_from_changeset"""
        request_list = list()

        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)

        if stub:
            p = stub
            p.id = changeset["id"]
            p.username = changeset["username"]
            p.email = changeset["email"]
            p.crypt_public = changeset["crypt_public"]
            p.sign_public = changeset["sign_public"]
            p.modified = modified_dt
            p._stub = False
        else:
            p = Persona(
                id=changeset["id"],
                username=changeset["username"],
                email=changeset["email"],
                crypt_public=changeset["crypt_public"],
                sign_public=changeset["sign_public"],
                modified=modified_dt,
            )

        # Update profile
        profile = Starmap.query.get(changeset["profile_id"])
        if profile is None or profile.get_state() == -1:
            request_list.append({
                "type": "Starmap",
                "id": changeset["profile_id"],
                "author_id": update_recipient.id,
                "recipient_id": update_sender.id,
            })

        if profile is None:
            profile = Starmap(id=changeset["profile_id"])
            profile.state = -1

        p.profile = profile

        # Update index
        index = Starmap.query.get(changeset["index_id"])
        if index is None or index.get_state() == -1:
            request_list.append({
                "type": "Starmap",
                "id": changeset["index_id"],
                "author_id": update_recipient.id,
                "recipient_id": update_sender.id,
            })

        if index is None:
            index = Starmap(id=changeset["index_id"])
            index.state = -1

        p.index = index

        # Update contacts
        for contact in changeset["contacts"]:
            c = Persona.query.get(contact["id"])

            if c is None:
                c = Persona(id=contact["id"], _stub=True)
                request_list.append({
                    "type": "Persona",
                    "id": contact["id"],
                    "author_id": update_recipient.id,
                    "recipient_id": update_sender.id,
                })

            p.contacts.append(c)

        app.logger.info("Created {} from changeset, now requesting {} linked objects".format(
            p, len(request_list)))

        for req in request_list:
            request_objects.send(Persona.create_from_changeset, message=req)

        return p

    def update_from_changeset(self, changeset, update_sender=None, update_recipient=None):
        """See Serializable.update_from_changeset"""
        request_list = list()

        # Update modified
        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)
        self.modified = modified_dt

        # Update username
        self.username = changeset["username"]
        app.logger.info("Updated {}'s {}".format(self.username, "username"))

        # Update email
        self.email = changeset["email"]
        app.logger.info("Updated {}'s {}".format(self.username, "email"))

        # Update contacts
        updated_contacts = list()
        requested_contacts = list()

        # remove_contacts contains all old contacts at first, all current
        # contacts get then removed so that the remaining can get deleted
        remove_contacts = set(self.contacts)

        for contact in changeset["contacts"]:
            c = Persona.query.get(contact["id"])

            if c is None:
                c = Persona(id=contact["id"], _stub=True)
                request_list.append({
                    "type": "Persona",
                    "id": contact["id"],
                    "author_id": update_recipient.id,
                    "recipient_id": update_sender.id,
                })
                requested_contacts.append(c)
            else:
                updated_contacts.append(c)

                try:
                    remove_contacts.remove(c)
                except KeyError:
                    pass

            self.contacts.append(c)

        for contact in remove_contacts:
            self.contacts.remove(contact)

        app.logger.info("Updated {}'s contacts: {} added, {} removed, {} requested".format(
            self.username, len(updated_contacts), len(remove_contacts), len(requested_contacts)))

        # Update profile
        profile = Starmap.query.get(changeset["profile_id"])
        if profile is None or profile.get_state() == -1:
            request_list.append({
                "type": "Starmap",
                "id": changeset["profile_id"],
                "author_id": update_recipient.id,
                "recipient_id": update_sender.id,
            })
            app.logger.info("Requested {}'s {}".format(self.username, "profile starmap"))
        else:
            self.profile = profile
            app.logger.info("Updated {}'s {}".format(self.username, "profile starmap"))

        # Update index
        index = Starmap.query.get(changeset["index_id"])
        if index is None or index.get_state() == -1:
            request_list.append({
                "type": "Starmap",
                "id": changeset["index_id"],
                "author_id": update_recipient.id,
                "recipient_id": update_sender.id,
            })
            app.logger.info("Requested {}'s new {}".format(self.username, "index starmap"))
        else:
            self.index = index
            app.logger.info("Updated {}'s {}".format(self.username, "index starmap"))

        app.logger.info("Updated {} from changeset. Requesting {} objects.".format(self, len(request_list)))

        for req in request_list:
            request_objects.send(Persona.create_from_changeset, message=req)


class Oneup(Serializable, db.Model):
    """A 1up is a vote that signals interest in a Star"""

    __tablename__ = "oneup"
    id = db.Column(db.String(32), primary_key=True)
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    state = db.Column(db.Integer, default=0)

    author = db.relationship("Persona",
        backref=db.backref('oneups'),
        primaryjoin="Persona.id==Oneup.author_id")
    author_id = db.Column(db.String(32), db.ForeignKey('persona.id'))

    star_id = db.Column(db.String(32), db.ForeignKey('star.id'))

    def __repr__(self):
        return "<1up <Persona {}> -> <Star {}> ({})>".format(self.author_id[:6], self.star_id[:6], self.get_state())

    def get_state(self):
        """
        Return publishing state of this 1up.

        Returns:
            Integer:
                -1 -- (disabled)
                 0 -- (active)
                 1 -- (unknown author)
        """
        return ONEUP_STATES[self.state][0]

    def set_state(self, new_state):
        """
        Set the publishing state of this 1up

        Parameters:
            new_state (int) code of the new state as defined in nucleus.ONEUP_STATES
        """
        if not isinstance(new_state, int) or new_state not in ONEUP_STATES.keys():
            raise ValueError("{} ({}) is not a valid 1up state").format(
                new_state, type(new_state))
        else:
            self.state = new_state

t_star_vesicles = db.Table(
    'star_vesicles',
    db.Column('star_id', db.String(32), db.ForeignKey('star.id')),
    db.Column('vesicle_id', db.String(32), db.ForeignKey('vesicle.id'))
)


class Star(Serializable, db.Model):
    """A Star represents a post"""

    __tablename__ = "star"

    _insert_required = ["id", "text", "created", "modified", "author_id", "planets"]
    _update_required = ["id", "text", "modified"]

    id = db.Column(db.String(32), primary_key=True)
    text = db.Column(db.Text)

    created = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())

    state = db.Column(db.Integer, default=0)

    author = db.relationship('Persona',
        backref=db.backref('starmap'),
        primaryjoin="Persona.id==Star.author_id")
    author_id = db.Column(db.String(32), db.ForeignKey('persona.id'))

    oneups = db.relationship('Oneup',
        backref='star',
        lazy='dynamic')

    planets = db.relationship('Planet',
        secondary='satellites',
        backref=db.backref('stars'),
        primaryjoin="satellites.c.star_id==star.c.id",
        secondaryjoin="satellites.c.planet_id==planet.c.id")

    vesicles = db.relationship(
        'Vesicle',
        secondary='star_vesicles',
        primaryjoin='star_vesicles.c.star_id==star.c.id',
        secondaryjoin='star_vesicles.c.vesicle_id==vesicle.c.id')

    def __repr__(self):
        try:
            ascii_text = self.text.encode('utf-8')
        except AttributeError:
            ascii_text = "No text content"
        return "<Star {}: {}>".format(
            self.id[:6],
            (ascii_text[:24] if len(ascii_text) <= 24 else ascii_text[:22] + ".."))

    def authorize(self, action, author_id=None):
        """Return True if this Star authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        if Serializable.authorize(self, action, author_id=author_id):
            return author_id == self.author.id
        return False

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """See Serializable.create_from_changeset"""
        created_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)
        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)

        if stub is not None:
            star = stub
            star.text = changeset["text"]
            star.author = None
            star.created = created_dt
            star.modified = modified_dt
        else:
            star = Star(
                id=changeset["id"],
                text=changeset["text"],
                author=None,
                created=created_dt,
                modified=modified_dt,
            )

        author = Persona.query.get(changeset["author_id"])
        if author is None:
            # TODO: Send request for author
            star.author_id = changeset["author_id"]
        else:
            star.author = author

        app.logger.info("Created new Star from changeset")

        return star

    def update_from_changeset(self, changeset, update_sender=None, update_recipient=None):
        """Update a Star from a changeset (See Serializable.update_from_changeset)"""
        # Update modified
        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)
        self.modified = modified_dt

        # Update text
        self.text = changeset["text"]

        app.logger.info("Updated {} from changeset".format(self))

    def export(self, update=False):
        """See Serializable.export"""

        data = Serializable.export(self, exclude=["planets", ], update=update)

        data["planets"] = list()
        for planet in self.planets:
            data["planets"].append(planet.export())

        return data

    def get_state(self):
        """
        Return publishing state of this star.

        Returns:
            Integer:
                -2 -- deleted
                -1 -- unavailable
                0 -- published
                1 -- draft
                2 -- private
                3 -- updating
        """
        return STAR_STATES[self.state][0]

    def set_state(self, new_state):
        """
        Set the publishing state of this star

        Parameters:
            new_state (int) code of the new state as defined in nucleus.STAR_STATES
        """
        if not isinstance(new_state, int) or new_state not in STAR_STATES.keys():
            raise ValueError("{} ({}) is not a valid star state").format(
                new_state, type(new_state))
        else:
            self.state = new_state

    def get_absolute_url(self):
        return url_for('star', id=self.id)

    def hot(self):
        """i reddit"""
        from math import log
        # Uncomment to assign a score with analytics.score
        #s = score(self)
        s = 1.0
        order = log(max(abs(s), 1), 10)
        sign = 1 if s > 0 else -1 if s < 0 else 0
        return round(order + sign * epoch_seconds(self.created) / 45000, 7)

    def oneupped(self):
        """
        Return True if active Persona has 1upped this Star
        """
        active_persona = Persona.query.get(session["active_persona"])
        oneup = self.oneups.filter_by(author=active_persona).first()
        if oneup is None or oneup.state < 0:
            return False
        else:
            return True

    def oneup_count(self):
        """
        Return the number of verified upvotes this Star has receieved

        Returns:
            Int: Number of upvotes
        """
        return self.oneups.filter_by(state=0).paginate(1).total

    def toggle_oneup(self, author_id=None):
        """
        Toggle 1up for this Star on/off

        Args:
            author_id (String): Optional Persona ID that issued the 1up. Defaults to active Persona.

        Returns:
            Oneup: The toggled oneup object

        Raises:
            PersonaNotFoundError: 1up author not found
            UnauthorizedError: Author is a foreign Persona
        """
        if author_id is None:
            author = Persona.query.get(session['active_persona'])
        else:
            author = Persona.query.get(author_id)

        if author is None:
            raise PersonaNotFoundError("1up author not found")

        if not author.controlled():
            raise UnauthorizedError("Can't toggle 1ups with foreign Persona {}".format(author))

        # Check whether 1up has been previously issued
        oneup = self.oneups.filter_by(author=author).first()
        if oneup is not None:
            old_state = oneup.get_state()
            oneup.set_state(-1) if oneup.state == 0 else oneup.set_state(0)
        else:
            old_state = False
            oneup = Oneup(id=uuid4().hex, star=self, author=author)

        # Commit 1up
        db.session.add(oneup)
        db.session.commit()
        app.logger.info("{verb} {obj}".format(verb="Toggled" if old_state else "Added", obj=oneup, ))

        return oneup


t_satellites = db.Table(
    'satellites',
    db.Column('star_id', db.String(32), db.ForeignKey('star.id')),
    db.Column('planet_id', db.String(32), db.ForeignKey('planet.id'))
)

t_planet_vesicles = db.Table(
    'planet_vesicles',
    db.Column('planet_id', db.String(32), db.ForeignKey('planet.id')),
    db.Column('vesicle_id', db.String(32), db.ForeignKey('vesicle.id'))
)


class Planet(Serializable, db.Model):
    """A Planet represents an attachment"""

    __tablename__ = 'planet'

    _insert_required = ["id", "title", "created", "modified", "source"]
    _update_required = ["id", "title", "modified", "source"]

    id = db.Column(db.String(32), primary_key=True)
    title = db.Column(db.Text)
    kind = db.Column(db.String(32))
    created = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    source = db.Column(db.String(128))
    state = db.Column(db.Integer, default=0)

    vesicles = db.relationship(
        'Vesicle',
        secondary='planet_vesicles',
        primaryjoin='planet_vesicles.c.planet_id==planet.c.id',
        secondaryjoin='planet_vesicles.c.vesicle_id==vesicle.c.id')

    __mapper_args__ = {
        'polymorphic_identity': 'planet',
        'polymorphic_on': kind
    }

    def __repr__(self):
        return "<Planet:{} [{}]>".format(self.kind, self.id[:6])

    def get_state(self):
        """
        Return publishing state of this planet.

        Returns:
            Integer:
                -2 -- deleted
                -1 -- unavailable
                0 -- published
                1 -- draft
                2 -- private
                3 -- updating
        """
        return PLANET_STATES[self.state][0]

    def set_state(self, new_state):
        """
        Set the publishing state of this planet

        Parameters:
            new_state (int) code of the new state as defined in nucleus.PLANET_STATES
        """
        if not isinstance(new_state, int) or new_state not in PLANET_STATES.keys():
            raise ValueError("{} ({}) is not a valid planet state").format(
                new_state, type(new_state))
        else:
            self.state = new_state

    def export(self, update=False):
        return Serializable.export(self, update=update)

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new Planet object from a changeset (See Serializable.create_from_changeset). """
        raise NotImplementedError

    def update_from_changeset(changeset, update_sender=None, update_recipient=None):
        """Update a new Planet object from a changeset (See Serializable.update_from_changeset). """
        raise NotImplementedError


class PicturePlanet(Planet):
    """A Picture attachment"""

    _insert_required = ["id", "title", "created", "modified", "source", "filename"]
    _update_required = ["id", "title", "modified", "source", "filename"]

    id = db.Column(db.String(32), ForeignKey('planet.id'), primary_key=True)
    filename = db.Column(db.Text)

    __mapper_args__ = {
        'polymorphic_identity': 'picture'
    }

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new Planet object from a changeset (See Serializable.create_from_changeset). """
        raise NotImplementedError

    def update_from_changeset(changeset, update_sender=None, update_recipient=None):
        """Update a new Planet object from a changeset (See Serializable.update_from_changeset). """
        raise NotImplementedError


class LinkPlanet(Planet):
    """A URL attachment"""

    _insert_required = ["id", "title", "kind", "created", "modified", "source", "url"]
    _update_required = ["id", "title", "modified", "source", "url"]

    id = db.Column(db.String(32), ForeignKey('planet.id'), primary_key=True)
    url = db.Column(db.Text)

    __mapper_args__ = {
        'polymorphic_identity': 'link'
    }

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new Planet object from a changeset (See Serializable.create_from_changeset). """
        raise NotImplementedError

    def update_from_changeset(changeset, update_sender=None, update_recipient=None):
        """Update a new Planet object from a changeset (See Serializable.update_from_changeset). """
        raise NotImplementedError


class Souma(Serializable, db.Model):
    """A physical machine in the Souma network"""

    __tablename__ = "souma"

    _insert_required = ["id", "modified", "crypt_public", "sign_public", "starmap_id"]
    id = db.Column(db.String(32), primary_key=True)

    crypt_private = db.Column(db.Text)
    crypt_public = db.Column(db.Text)
    sign_private = db.Column(db.Text)
    sign_public = db.Column(db.Text)

    starmap_id = db.Column(db.String(32), db.ForeignKey('starmap.id'))
    starmap = db.relationship('Starmap')

    def __str__(self):
        return "<Souma [{}]>".format(self.id[:6])

    def authorize(self, action, author_id=None):
        """Return True if this Souma authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        return False

    def generate_keys(self):
        """ Generate new RSA keypairs for signing and encrypting. Commit to DB afterwards! """

        # TODO: Store keys encrypted
        rsa1 = RsaPrivateKey.Generate()
        self.sign_private = str(rsa1)
        self.sign_public = str(rsa1.public_key)

        rsa2 = RsaPrivateKey.Generate()
        self.crypt_private = str(rsa2)
        self.crypt_public = str(rsa2.public_key)

    def encrypt(self, data):
        """ Encrypt data using RSA """

        if self.crypt_public == "":
            raise ValueError("Error encrypting: No public encryption key found for {}".format(self))

        key_public = RsaPublicKey.Read(self.crypt_public)
        return key_public.Encrypt(data)

    def decrypt(self, cypher):
        """ Decrypt cyphertext using RSA """

        if self.crypt_private == "":
            raise ValueError("Error decrypting: No private encryption key found for {}".format(self))

        key_private = RsaPrivateKey.Read(self.crypt_private)
        return key_private.Decrypt(cypher)

    def sign(self, data):
        """ Sign data using RSA """
        from base64 import urlsafe_b64encode

        if self.sign_private == "":
            raise ValueError("Error signing: No private signing key found for {}".format(self))

        key_private = RsaPrivateKey.Read(self.sign_private)
        signature = key_private.Sign(data)
        return urlsafe_b64encode(signature)

    def verify(self, data, signature_b64):
        """ Verify a signature using RSA """
        from base64 import urlsafe_b64decode

        if self.sign_public == "":
            raise ValueError("Error verifying: No public signing key found for {}".format(self))

        signature = urlsafe_b64decode(signature_b64)
        key_public = RsaPublicKey.Read(self.sign_public)
        return key_public.Verify(data, signature)

t_starmap = db.Table(
    'starmap_index',
    db.Column('starmap_id', db.String(32), db.ForeignKey('starmap.id')),
    db.Column('star_id', db.String(32), db.ForeignKey('star.id'))
)

t_starmap_vesicles = db.Table(
    'starmap_vesicles',
    db.Column('starmap_id', db.String(32), db.ForeignKey('starmap.id')),
    db.Column('vesicle_id', db.String(32), db.ForeignKey('vesicle.id'))
)


class Starmap(Serializable, db.Model):
    """
    Starmaps are collections of objects with associated layout information.

    Atributes:
        id: 32 byte ID generated by uuid4().hex
        modified: Datetime of last recorded modification
        author: Persona that created this Starmap
        kind: For what kind of context is this Starmap used
        index: List of Stars that are contained in this Starmap
        vesicles: List of Vesicles that describe this Starmap
    """
    __tablename__ = 'starmap'

    _insert_required = ["id", "modified", "author_id", "kind", "state"]
    _update_required = ["id", "modified", "index"]

    id = db.Column(db.String(32), primary_key=True)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    kind = db.Column(db.String(16))
    state = db.Column(db.Integer, default=0)

    author_id = db.Column(
        db.String(32),
        db.ForeignKey('persona.id', use_alter=True, name="fk_author_id"))
    author = db.relationship('Persona',
        backref=db.backref('starmaps'),
        primaryjoin="Persona.id==Starmap.author_id",
        post_update=True)

    index = db.relationship(
        'Star',
        secondary='starmap_index',
        backref="starmaps",
        lazy="dynamic",
        primaryjoin='starmap_index.c.starmap_id==starmap.c.id',
        secondaryjoin='starmap_index.c.star_id==star.c.id')

    vesicles = db.relationship(
        'Vesicle',
        secondary='starmap_vesicles',
        primaryjoin='starmap_vesicles.c.starmap_id==starmap.c.id',
        secondaryjoin='starmap_vesicles.c.vesicle_id==vesicle.c.id')

    def __contains__(self, key):
        """Return True if the given key is contained in this Starmap.

        Args:
            key: db.model.key to look for
        """
        return (key in self.index)

    def __repr__(self):
        if self.kind == "persona_profile":
            name = "Persona-Profile"
        elif self.kind == "group_profile":
            name = "Group-Profile"
        else:
            name = "Starmap"

        return "<{} (by {}) [{}]>".format(name, self.author, self.id[:6])

    def __len__(self):
        return self.index.paginate(1).total

    def authorize(self, action, author_id=None):
        """Return True if this Starmap authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        if Serializable.authorize(self, action, author_id=author_id):
            if self.kind == "persona_profile":
                # p = Persona.query.filter(Persona.profile_id == self.id).first()
                # return p.id == author_id
                return self.author.id == author_id
            elif self.kind == "group_profile":
                # Everyone can update
                if action == "update":
                    return True
                # Only author can insert and delete
                elif self.author_id == author_id:
                    return True
            elif self.kind == "index":
                p = Persona.query.filter(Persona.index_id == self.id)
                return p.id == author_id
        return False

    def get_state(self):
        """
        Return publishing state of this star.

        Returns:
            Integer:
                -2 -- deleted
                -1 -- unavailable
                0 -- published
                1 -- draft
                2 -- private
                3 -- updating
        """
        return STAR_STATES[self.state][0]

    def set_state(self, new_state):
        """
        Set the publishing state of this star

        Parameters:
            new_state (int) code of the new state as defined in nucleus.STAR_STATES
        """
        if not isinstance(new_state, int) or new_state not in STAR_STATES.keys():
            raise ValueError("{} ({}) is not a valid star state").format(
                new_state, type(new_state))
        else:
            self.state = new_state

    def get_absolute_url(self):
        """Return URL for this Starmap depending on kind"""
        # import pdb; pdb.set_trace()
        if self.kind == "persona_profile":
            p = Persona.query.filter(Persona.profile_id == self.id).first()
            return url_for("persona", id=p.id)
        elif self.kind == "group_profile":
            g = Group.query.filter(Group.profile_id == self.id).first()
            return url_for("group", id=g.id)
        elif self.kind == "index":
            p = Persona.query.filter(Persona.index_id == self.id).first()
            return url_for("persona", id=p.id)

    def export(self, update=False):
        data = Serializable.export(self, exclude=["index", ], update=update)

        data["index"] = list()
        for star in self.index.filter('Star.state >= 0'):
            data["index"].append({
                "id": star.id,
                "modified": star.modified.isoformat(),
                "author_id": star.author.id
            })

        return data

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new Starmap object from a changeset

        Args:
            changeset (dict): Contains all keys from self._insert_required

        Returns:
            Starmap: The new object

        Raises:
            ValueError: If a value is invalid
            KeyError: If a required Value is missing
        """
        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)

        author = Persona.query.get(changeset["author_id"])
        if author is None:
            raise PersonaNotFoundError("Starmap author not known")

        if stub is not None:
            new_starmap = stub
            new_starmap.modified = modified_dt
            new_starmap.author = author
            new_starmap.kind = changeset["kind"]
        else:
            new_starmap = Starmap(
                id=changeset["id"],
                modified=modified_dt,
                author=author,
                kind=changeset["kind"]
            )

        request_list = list()
        for star_changeset in changeset["index"]:
            star = Star.query.get(star_changeset["id"])
            star_changeset_modified = iso8601.parse_date(star_changeset["modified"]).replace(tzinfo=None)

            if star is None or star.get_state() == -1 or star.modified < star_changeset_modified:
                request_list.append({
                    "type": "Star",
                    "id": star_changeset["id"],
                    "author_id": update_recipient.id,
                    "recipient_id": update_sender.id,
                })

                if star is None:
                    star_author = Persona.query.get(star_changeset["author_id"])
                    if star_author is not None:
                        star = Star(
                            id=star_changeset["id"],
                            modified=star_changeset_modified,
                            author=star_author
                        )
                        star.set_state(-1)
                        db.session.add(star)
                        db.session.commit()

            new_starmap.index.append(star)

        db.session.add(new_starmap)
        db.session.commit()

        for req in request_list:
            request_objects.send(Starmap.create_from_changeset, message=req)

        return new_starmap

    def update_from_changeset(self, changeset, update_sender=None, update_recipient=None):
        """Update the Starmap's index using a changeset

        Args:
            changeset (dict): Contains a key for every attribute to update

        Raises:
            ValueError: If a value in the changeset is invalid
        """
        # Update modified
        modified = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)
        self.modified = modified

        # Update index
        remove_stars = set([s.id for s in self.index if s is not None])
        added_stars = list()
        request_list = list()
        for star_changeset in changeset["index"]:
            star = Star.query.get(star_changeset["id"])
            star_changeset_modified = iso8601.parse_date(star_changeset["modified"]).replace(tzinfo=None)

            if star is not None and star.id in remove_stars:
                remove_stars.remove(star.id)

            if star is None or star.get_state() == -1 or star.modified < star_changeset_modified:
                # No copy of Star available or copy is outdated

                request_list.append({
                    "type": "Star",
                    "id": star_changeset["id"],
                    "author_id": update_recipient.id,
                    "recipient_id": update_sender.id,
                })

                if star is None:
                    star_author = Persona.query.get(star_changeset["author_id"])
                    if star_author is not None:
                        star = Star(
                            id=star_changeset["id"],
                            modified=star_changeset_modified,
                            author=star_author
                        )
                        star.set_state(-1)
                        db.session.add(star)
                        db.session.commit()

            self.index.append(star)
            added_stars.append(star)

        for s_id in remove_stars:
            s = Star.query.get(s_id)
            self.index.remove(s)

        app.logger.info("Updated {}: {} stars added, {} requested, {} removed".format(
            self, len(added_stars), len(request_list), len(remove_stars)))

        for req in request_list:
            request_objects.send(Starmap.create_from_changeset, message=req)

t_group_vesicles = db.Table(
    'group_vesicles',
    db.Column('group_id', db.String(32), db.ForeignKey('group.id')),
    db.Column('vesicle_id', db.String(32), db.ForeignKey('vesicle.id'))
)


class Group(Serializable, db.Model):
    """
        Represents an entity that is comprised of users collaborating on
        stars
    """

    __tablename__ = "group"
    _insert_required = ["id", "modified", "author_id", "groupname", "description", "profile_id", "state"]
    _update_required = ["id", "modified", "state"]

    id = db.Column(db.String(32), primary_key=True)
    modified = db.Column(db.DateTime, default=datetime.datetime.utcnow())
    groupname = db.Column(db.String(80))
    description = db.Column(db.Text)

    state = db.Column(db.Integer, default=0)

    author_id = db.Column(db.String(32), db.ForeignKey('persona.id'))
    author = db.relationship("Persona", primaryjoin="persona.c.id==group.c.author_id")

    profile_id = db.Column(db.String(32), db.ForeignKey('starmap.id'))
    profile = db.relationship('Starmap', primaryjoin='starmap.c.id==group.c.profile_id')

    vesicles = db.relationship(
        'Vesicle',
        secondary='group_vesicles',
        primaryjoin='group_vesicles.c.group_id==group.c.id',
        secondaryjoin='group_vesicles.c.vesicle_id==vesicle.c.id')

    # Make this work if needed!
    """
    members = db.relationship(
        "Persona",
        backref="groups",
        primaryjoin='group.c.id==persona.c.?????_id' # TODO:How to HBTM?!
    )"""

    def __repr__(self):
        try:
            name = " {} ".format(self.groupname.encode('utf-8'))
        except AttributeError:
            name = ""
        return "<Group{}[{}]>".format(name, self.id[:6])

    def authorize(self, action, author_id=None):
        """Return True if this Group authorizes `action` for `author_id`

        Args:
            action (String): Action to be performed (see Synapse.CHANGE_TYPES)
            author_id (String): Persona ID that wants to perform the action

        Returns:
            Boolean: True if authorized
        """
        if Serializable.authorize(self, action, author_id=author_id):
            return self.author_id == author_id
        return False

    def get_state(self):
        """
        Return publishing state of this Group. (temporarily uses planet states)

        Returns:
            Integer:
                -2 -- deleted
                -1 -- unavailable
                0 -- published
                1 -- draft
                2 -- private
                3 -- updating
        """
        return PLANET_STATES[self.state][0]

    def set_state(self, new_state):
        """
        Set the publishing state of this Group (temporarily uses planet states)

        Parameters:
            new_state (int) code of the new state as defined in nucleus.PLANET_STATES
        """
        if not isinstance(new_state, int) or new_state not in PLANET_STATES.keys():
            raise ValueError("{} ({}) is not a valid planet state").format(
                new_state, type(new_state))
        else:
            self.state = new_state

    @staticmethod
    def create_from_changeset(changeset, stub=None, update_sender=None, update_recipient=None):
        """Create a new group from changeset"""
        modified_dt = iso8601.parse_date(changeset["modified"]).replace(tzinfo=None)

        if stub is not None:
            group = stub
            group.id = changeset["id"]
            group.author = None
            group.profile = None
            group.modified = modified_dt
            group.groupname = changeset["groupname"]
            group.description = changeset["description"]
        else:
            group = Group(
                id=changeset["id"],
                modified=modified_dt,
                author=None,
                profile=None,
                groupname=changeset["groupname"],
                description=changeset["description"]
            )

        group.set_state(int(changeset["state"]))
        request_list = list()

        # Update profile
        profile = Starmap.query.get(changeset["profile_id"])
        if profile is None or profile.get_state() == -1:
            request_list.append({
                "type": "Starmap",
                "id": changeset["profile_id"],
                "author_id": update_recipient.id if update_recipient else None,
                "recipient_id": update_sender.id if update_sender else None,
            })

        if profile is None:
            profile = Starmap(
                id=changeset["profile_id"],
                kind="group_profile"
            )
            profile.state = -1

        group.profile = profile

        # Update author
        author = Persona.query.get(changeset["author_id"])
        if author is None or author._stub:
            request_list.append({
                "type": "Persona",
                "id": changeset["author_id"],
                "author_id": update_recipient.id if update_recipient else None,
                "recipient_id": update_sender.id if update_sender else None,
            })

        if author is None:
            author = Persona(
                id=changeset["author_id"],
            )
            author._stub = True

        group.author = author
        return group
